import argparse
import logging
import os
from typing import Tuple
import warnings

import torch
import torchvision

from .. import headmeta
from ..configurable import Configurable
from . import basenetworks, heads, model_migration, nets, tracking_heads
from . import swin_transformer, xcit
from .tracking_base import TrackingBase


# monkey patch torchvision for mobilenetv2 checkpoint backwards compatibility
if hasattr(torchvision.models, 'mobilenetv2') \
   and not hasattr(torchvision.models.mobilenet, 'ConvBNReLU'):
    torchvision.models.mobilenet.ConvBNReLU = torchvision.models.mobilenetv2.ConvBNReLU
    torchvision.models.mobilenet.InvertedResidual = torchvision.models.mobilenetv2.InvertedResidual

# monkey patch torchvision for mobilenetv3 checkpoint backwards compatibility
if hasattr(torchvision.models, 'mobilenetv3') \
   and not hasattr(torchvision.ops.misc.SqueezeExcitation, 'avgpool'):
    torchvision.ops.misc.SqueezeExcitation.avgpool = torch.nn.AdaptiveAvgPool2d(1)
    torchvision.ops.misc.SqueezeExcitation.activation = torch.nn.ReLU()
    torchvision.ops.misc.SqueezeExcitation.scale_activation = torch.nn.Hardsigmoid()


# generate hash values with: shasum -a 256 filename.pkl

PRETRAINED_UNAVAILABLE = object()

# Dataset cocokp is implied. All other datasets need to be explicit.
# Use http instead of https to avoid SSL certificate issues on Windows.
CHECKPOINT_URLS = {}

BASE_TYPES = set([
    basenetworks.MobileNetV2,
    basenetworks.MobileNetV3,
    basenetworks.Resnet,
    basenetworks.ShuffleNetV2,
    basenetworks.ShuffleNetV2K,
    basenetworks.SqueezeNet,
    basenetworks.SwinTransformer,
    basenetworks.XCiT,
    basenetworks.EffNetV2,
    TrackingBase,
])
BASE_FACTORIES = {
    'mobilenetv2': lambda: basenetworks.MobileNetV2('mobilenetv2', torchvision.models.mobilenet_v2),
    'mobilenetv3large': lambda: basenetworks.MobileNetV3(
        'mobilenetv3large', torchvision.models.mobilenet_v3_large),
    'mobilenetv3small': lambda: basenetworks.MobileNetV3(
        'mobilenetv3small', torchvision.models.mobilenet_v3_small, 576),
    'resnet18': lambda: basenetworks.Resnet('resnet18', torchvision.models.resnet18, 512),
    'resnet50': lambda: basenetworks.Resnet('resnet50', torchvision.models.resnet50),
    'resnet101': lambda: basenetworks.Resnet('resnet101', torchvision.models.resnet101),
    'resnet152': lambda: basenetworks.Resnet('resnet152', torchvision.models.resnet152),
    'resnext50': lambda: basenetworks.Resnet('resnext50', torchvision.models.resnext50_32x4d),
    'resnext101': lambda: basenetworks.Resnet('resnext101', torchvision.models.resnext101_32x8d),
    'shufflenetv2x1': lambda: basenetworks.ShuffleNetV2(
        'shufflenetv2x1', torchvision.models.shufflenet_v2_x1_0, 1024),
    'shufflenetv2x2': lambda: basenetworks.ShuffleNetV2(
        # defined in torchvision as [4, 8, 4], [24, 244, 488, 976, 2048]
        'shufflenetv2x2', torchvision.models.shufflenet_v2_x2_0),
    'shufflenetv2k16': lambda: basenetworks.ShuffleNetV2K(
        'shufflenetv2k16', [4, 8, 4], [24, 348, 696, 1392, 1392]),
    'shufflenetv2k20': lambda: basenetworks.ShuffleNetV2K(
        'shufflenetv2k20', [5, 10, 5], [32, 512, 1024, 2048, 2048]),
    'shufflenetv2kx5': lambda: basenetworks.ShuffleNetV2K(
        'shufflenetv2kx5', [6, 13, 6], [42, 640, 1280, 2560, 2560]),
    'shufflenetv2k30': lambda: basenetworks.ShuffleNetV2K(
        'shufflenetv2k30', [8, 16, 6], [32, 512, 1024, 2048, 2048]),
    'shufflenetv2k44': lambda: basenetworks.ShuffleNetV2K(
        'shufflenetv2k44', [12, 24, 8], [32, 512, 1024, 2048, 2048]),
    'squeezenet': lambda: basenetworks.SqueezeNet('squeezenet', torchvision.models.squeezenet1_1),
    # Swin architectures: swin_t is roughly equivalent to unmodified resnet50
    'swin_t': lambda: basenetworks.SwinTransformer(
        'swin_t', swin_transformer.swin_tiny_patch4_window7),
    'swin_s': lambda: basenetworks.SwinTransformer(
        'swin_s', swin_transformer.swin_small_patch4_window7),
    'swin_b': lambda: basenetworks.SwinTransformer(
        'swin_b', swin_transformer.swin_base_patch4_window7),
    'swin_b_window_12': lambda: basenetworks.SwinTransformer(
        'swin_b_window_12', swin_transformer.swin_base_patch4_window12),
    'swin_l': lambda: basenetworks.SwinTransformer(
        'swin_l', swin_transformer.swin_large_patch4_window7),
    'swin_l_window_12': lambda: basenetworks.SwinTransformer(
        'swin_l_window_12', swin_transformer.swin_large_patch4_window12),
    # XCiT architectures: xcit_small_12_p16 is roughly equivalent to unmodified resnet50
    'xcit_nano_12_p16': lambda: basenetworks.XCiT('xcit_nano_12_p16', xcit.xcit_nano_12_p16),
    'xcit_tiny_12_p16': lambda: basenetworks.XCiT('xcit_tiny_12_p16', xcit.xcit_tiny_12_p16),
    'xcit_tiny_24_p16': lambda: basenetworks.XCiT('xcit_tiny_24_p16', xcit.xcit_tiny_24_p16),
    'xcit_small_12_p16': lambda: basenetworks.XCiT('xcit_small_12_p16', xcit.xcit_small_12_p16),
    'xcit_small_24_p16': lambda: basenetworks.XCiT('xcit_small_24_p16', xcit.xcit_small_24_p16),
    'xcit_medium_24_p16': lambda: basenetworks.XCiT('xcit_medium_24_p16', xcit.xcit_medium_24_p16),
    'xcit_large_24_p16': lambda: basenetworks.XCiT('xcit_large_24_p16', xcit.xcit_large_24_p16),
    'xcit_nano_12_p8': lambda: basenetworks.XCiT('xcit_nano_12_p8', xcit.xcit_nano_12_p8),
    'xcit_tiny_12_p8': lambda: basenetworks.XCiT('xcit_tiny_12_p8', xcit.xcit_tiny_12_p8),
    'xcit_tiny_24_p8': lambda: basenetworks.XCiT('xcit_tiny_24_p8', xcit.xcit_tiny_24_p8),
    'xcit_small_12_p8': lambda: basenetworks.XCiT('xcit_small_12_p8', xcit.xcit_small_12_p8),
    'xcit_small_24_p8': lambda: basenetworks.XCiT('xcit_small_24_p8', xcit.xcit_small_24_p8),
    'xcit_medium_24_p8': lambda: basenetworks.XCiT('xcit_medium_24_p8', xcit.xcit_medium_24_p8),
    'xcit_large_24_p8': lambda: basenetworks.XCiT('xcit_large_24_p8', xcit.xcit_large_24_p8),
    # Parameters for the EffNetV2 construction
    # expansion ratio, channels, number of layers of this type, stride, use squeeze+excitation
    # t, c, n, s, SE
    'effnetv2_s': lambda: basenetworks.EffNetV2('effnetv2_s',
                                                [
                                                    [1, 24, 2, 1, 0],
                                                    [4, 48, 4, 2, 0],
                                                    [4, 64, 4, 2, 0],
                                                    [4, 128, 6, 2, 1],
                                                    [6, 160, 9, 1, 1],
                                                    [6, 256, 15, 2, 1],
                                                ],
                                                stride=32),
    'effnetv2_m': lambda: basenetworks.EffNetV2('effnetv2_m',
                                                [
                                                    [1, 24, 3, 1, 0],
                                                    [4, 48, 5, 2, 0],
                                                    [4, 80, 5, 2, 0],
                                                    [4, 160, 7, 2, 1],
                                                    [6, 176, 14, 1, 1],
                                                    [6, 304, 18, 2, 1],
                                                    [6, 512, 5, 1, 1],
                                                ],
                                                stride=32),
    'effnetv2_l': lambda: basenetworks.EffNetV2('effnetv2_l',
                                                [
                                                    [1, 32, 4, 1, 0],
                                                    [4, 64, 7, 2, 0],
                                                    [4, 96, 7, 2, 0],
                                                    [4, 192, 10, 2, 1],
                                                    [6, 224, 19, 1, 1],
                                                    [6, 384, 25, 2, 1],
                                                    [6, 640, 7, 1, 1],
                                                ],
                                                stride=32),
    'effnetv2_xl': lambda: basenetworks.EffNetV2('effnetv2_xl',
                                                 [
                                                     [1, 32, 4, 1, 0],
                                                     [4, 64, 8, 2, 0],
                                                     [4, 96, 8, 2, 0],
                                                     [4, 192, 16, 2, 1],
                                                     [6, 256, 24, 1, 1],
                                                     [6, 512, 32, 2, 1],
                                                     [6, 640, 8, 1, 1],
                                                 ],
                                                 stride=32),
    'effnetv2_s16_s': lambda: basenetworks.EffNetV2('effnetv2_s16_s',
                                                    [
                                                        [1, 24, 2, 1, 0],
                                                        [4, 48, 4, 2, 0],
                                                        [4, 64, 4, 2, 0],
                                                        [4, 128, 6, 2, 1],
                                                        [6, 160, 9, 1, 1],
                                                        # [6, 256, 15, -1, 1],  # -1 = dilated con
                                                    ],
                                                    stride=16),
    'effnetv2_s16_m': lambda: basenetworks.EffNetV2('effnetv2_s16_m',
                                                    [
                                                        [1, 24, 3, 1, 0],
                                                        [4, 48, 5, 2, 0],
                                                        [4, 80, 5, 2, 0],
                                                        [4, 160, 7, 2, 1],
                                                        [6, 176, 14, 1, 1],
                                                        # [6, 304, 18, -1, 1], # -1 = dilated conv
                                                        # [6, 512, 5, 1, 1],
                                                    ],
                                                    stride=16),
    'effnetv2_s16_l': lambda: basenetworks.EffNetV2('effnetv2_s16_l',
                                                    [
                                                        [1, 32, 4, 1, 0],
                                                        [4, 64, 7, 2, 0],
                                                        [4, 96, 7, 2, 0],
                                                        [4, 192, 10, 2, 1],
                                                        [6, 224, 19, 1, 1],
                                                        # [6, 384, 25, -1, 1],  # -1=dilated conv
                                                        # [6, 640, 7, 1, 1],
                                                    ],
                                                    stride=16),
    'effnetv2_s16_xl': lambda: basenetworks.EffNetV2('effnetv2_s16_xl',
                                                     [
                                                         [1, 32, 4, 1, 0],
                                                         [4, 64, 8, 2, 0],
                                                         [4, 96, 8, 2, 0],
                                                         [4, 192, 16, 2, 1],
                                                         [6, 256, 24, 1, 1],
                                                         # [6, 512, 32, -1, 1],  # -1 = dilated c
                                                         # [6, 640, 8, 1, 1],
                                                     ],
                                                     stride=16),
    'botnet': lambda: basenetworks.BotNet('botnet'),
}
# base factories that wrap other base factories:
BASE_FACTORIES['tshufflenetv2k16'] = lambda: TrackingBase(BASE_FACTORIES['shufflenetv2k16']())
BASE_FACTORIES['tshufflenetv2k30'] = lambda: TrackingBase(BASE_FACTORIES['shufflenetv2k30']())
BASE_FACTORIES['tresnet50'] = lambda: TrackingBase(BASE_FACTORIES['resnet50']())

#: headmeta class to head class
HEADS = {
    headmeta.Cif: heads.CompositeField4,
    headmeta.Caf: heads.CompositeField4,
    headmeta.CifDet: heads.CompositeField4,
    headmeta.TSingleImageCif: tracking_heads.TBaseSingleImage,
    headmeta.TSingleImageCaf: tracking_heads.TBaseSingleImage,
    headmeta.Tcaf: tracking_heads.Tcaf,
}

LOG = logging.getLogger(__name__)


def local_checkpoint_path(checkpoint):
    if os.path.exists(checkpoint):
        return checkpoint

    if checkpoint in CHECKPOINT_URLS:
        url = CHECKPOINT_URLS[checkpoint]

        base_dir = None
        if hasattr(torch, 'hub') and hasattr(torch.hub, 'get_dir'):
            # new in pytorch 1.6.0
            base_dir = torch.hub.get_dir()
        elif os.getenv('TORCH_HOME'):
            base_dir = os.getenv('TORCH_HOME')
        elif os.getenv('XDG_CACHE_HOME'):
            base_dir = os.path.join(os.getenv('XDG_CACHE_HOME'), 'torch')
        else:
            base_dir = os.path.expanduser(os.path.join('~', '.cache', 'torch'))

        file_name = os.path.join(base_dir, 'checkpoints', os.path.basename(url))
        if os.path.exists(file_name):
            return file_name

    return None


class Factory(Configurable):
    base_name = None
    checkpoint = None
    cross_talk = 0.0
    download_progress = True
    head_consolidation = 'filter_and_extend'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.base_name is not None:
            assert self.checkpoint is None
        if self.checkpoint is not None:
            assert self.base_name is None

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        for bn in BASE_TYPES:
            bn.cli(parser)
        for hn in set(HEADS.values()):
            hn.cli(parser)

        group = parser.add_argument_group('network configuration')
        available_checkpoints = [n for n, url in CHECKPOINT_URLS.items()
                                 if url is not PRETRAINED_UNAVAILABLE]
        group.add_argument(
            '--checkpoint', default=cls.checkpoint,
            help=(
                'Path to a local checkpoint. '
                'Or provide one of the following to download a pretrained model: {}'
                ''.format(available_checkpoints)
            )
        )
        group.add_argument('--basenet', default=cls.base_name,
                           help='base network, one of {}'.format(list(BASE_FACTORIES.keys())))
        group.add_argument('--cross-talk', default=cls.cross_talk, type=float,
                           help='[experimental]')
        assert cls.download_progress
        group.add_argument('--no-download-progress', dest='download_progress',
                           default=True, action='store_false',
                           help='suppress model download progress bar')
        group.add_argument('--head-consolidation',
                           choices=('keep', 'create', 'filter_and_extend'),
                           default=cls.head_consolidation,
                           help=('consolidation strategy for a checkpoint\'s head '
                                 'networks and the heads specified by the datamodule'))

    @classmethod
    def configure(cls, args: argparse.Namespace):
        for bn in BASE_TYPES:
            bn.configure(args)
        for hn in set(HEADS.values()):
            hn.configure(args)

        cls.base_name = args.basenet
        cls.checkpoint = args.checkpoint
        cls.cross_talk = args.cross_talk
        cls.download_progress = args.download_progress
        cls.head_consolidation = args.head_consolidation

    def factory(self, *, head_metas=None) -> Tuple[nets.Shell, int]:
        if self.base_name:
            assert head_metas
            assert self.checkpoint is None
            net_cpu: nets.Shell = self.from_scratch(head_metas)
            net_cpu = self.init_net(net_cpu)
            epoch = 0
            return net_cpu, epoch

        net_cpu, epoch = self.from_checkpoint()
        if head_metas is not None:
            self.consolidate_heads(net_cpu, head_metas)

        net_cpu = self.init_net(net_cpu)
        return net_cpu, epoch

    def consolidate_heads(self, net_cpu, head_metas):
        if self.head_consolidation == 'keep':
            LOG.info('keeping heads from loaded checkpoint')
            # Match head metas by name and overwrite with meta from checkpoint.
            # This makes sure that the head metas have their head_index and
            # base_stride attributes set.
            input_head_meta_indices = {(meta.dataset, meta.name): i
                                       for i, meta in enumerate(head_metas)}
            for hn in net_cpu.head_nets:
                input_index = input_head_meta_indices.get((hn.meta.dataset, hn.meta.name), None)
                if input_index is None:
                    continue
                head_metas[input_index] = hn.meta
        elif self.head_consolidation == 'create':
            LOG.info('creating new heads')
            headnets = [HEADS[h.__class__](h, net_cpu.base_net.out_features)
                        for h in head_metas]
            net_cpu.set_head_nets(headnets)
        elif self.head_consolidation == 'filter_and_extend':
            LOG.info('filtering for dataset heads and extending existing heads')
            existing_headnets = {(hn.meta.dataset, hn.meta.name): hn
                                 for hn in net_cpu.head_nets}
            headnets = []
            for meta_i, meta in enumerate(head_metas):
                if (meta.dataset, meta.name) in existing_headnets:
                    hn = existing_headnets[(meta.dataset, meta.name)]
                    headnets.append(hn)
                    # Match head metas by name and overwrite with meta from checkpoint.
                    # This makes sure that the head metas have their head_index and
                    # base_stride attributes set.
                    head_metas[meta_i] = hn.meta
                else:
                    headnets.append(
                        HEADS[meta.__class__](meta, net_cpu.base_net.out_features))
            net_cpu.set_head_nets(headnets)
        else:
            raise Exception('head strategy {} unknown'.format(self.head_consolidation))

    def from_checkpoint(self) -> Tuple[nets.Shell, int]:
        checkpoint = self.checkpoint
        if not checkpoint:
            checkpoint = 'shufflenetv2k16'

        if CHECKPOINT_URLS.get(checkpoint, None) is PRETRAINED_UNAVAILABLE:
            raise Exception(
                'The pretrained model for "{}" is not available yet '
                'in this release cycle. Use one of {}.'.format(
                    checkpoint,
                    [k for k, v in CHECKPOINT_URLS.items() if v is not PRETRAINED_UNAVAILABLE],
                )
            )
        checkpoint = CHECKPOINT_URLS.get(checkpoint, checkpoint)

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=torch.serialization.SourceChangeWarning)

            if checkpoint.startswith('http'):
                checkpoint = torch.hub.load_state_dict_from_url(
                    checkpoint,
                    check_hash=not checkpoint.startswith('https'),
                    progress=self.download_progress)
            else:
                try:
                    checkpoint = torch.load(checkpoint)
                except FileNotFoundError as e:
                    raise Exception('Checkpoint "{}" not found. Pre-trained checkpoints: {}'.format(
                        checkpoint,
                        [k for k, v in CHECKPOINT_URLS.items() if v is not PRETRAINED_UNAVAILABLE]
                    )) from e

        net_cpu: nets.Shell = checkpoint['model']
        epoch = checkpoint['epoch']

        # normalize for backwards compatibility
        model_migration.model_migration(net_cpu)

        return net_cpu, epoch

    def from_scratch(self, head_metas) -> nets.Shell:
        if self.base_name not in BASE_FACTORIES:
            raise Exception('basenet {} unknown'.format(self.base_name))

        basenet = BASE_FACTORIES[self.base_name]()
        headnets = [HEADS[h.__class__](h, basenet.out_features) for h in head_metas]

        net_cpu = nets.Shell(basenet, headnets)
        nets.model_defaults(net_cpu)
        return net_cpu

    def init_net(self, net_cpu):
        if self.cross_talk:
            net_cpu.process_input = nets.CrossTalk(self.cross_talk)

        # initialize for eval
        net_cpu.eval()

        LOG.debug(net_cpu)
        return net_cpu



import openpifpaf

from .wholebody import Wholebody


def register():
    openpifpaf.DATAMODULES['wholebody'] = Wholebody
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k16-wholebody'] = 'http://github.com/DuncanZauss/' \
        'openpifpaf_assets/releases/download/v0.1.0/sk16_wholebody.pkl'
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k30-wholebody'] = 'http://github.com/DuncanZauss/' \
        'openpifpaf_assets/releases/download/v0.1.0/sk30_wholebody.pkl'



import argparse
import logging
import math

import torch

LOG = logging.getLogger(__name__)


class SoftClamp(torch.nn.Module):
    def __init__(self, max_value):
        super().__init__()
        self.max_value = max_value

    def forward(self, x):
        # Backprop rule pre-multiplies by input. Therefore, for a constant
        # gradient above the max bce threshold, need to divide by the input.
        # Just like gradient-clipping, but inline:
        above_max = x > self.max_value
        # x[above_max] /= th[above_max].detach() / self.max_value
        x[above_max] = self.max_value + torch.log(1 - self.max_value + x[above_max])

        return x


class Bce(torch.nn.Module):
    background_weight = 1.0
    focal_alpha = 0.5
    focal_gamma = 1.0
    focal_detach = False
    focal_clamp = True
    soft_clamp_value = 5.0

    # choose low value for force-complete-pose and Focal loss modification
    background_clamp = -15.0

    # 0.02 -> -3.9, 0.01 -> -4.6, 0.001 -> -7, 0.0001 -> -9
    min_bce = 0.0  # 1e-6 corresponds to x~=14, 1e-10 -> 20

    def __init__(self, **kwargs):
        super().__init__()
        for n, v in kwargs.items():
            assert hasattr(self, n)
            setattr(self, n, v)

        self.soft_clamp = None
        if self.soft_clamp_value:
            self.soft_clamp = SoftClamp(self.soft_clamp_value)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('Bce Loss')
        group.add_argument('--background-weight', default=cls.background_weight, type=float,
                           help='BCE weight where ground truth is background')
        group.add_argument('--focal-alpha', default=cls.focal_alpha, type=float,
                           help='scale parameter of focal loss')
        group.add_argument('--focal-gamma', default=cls.focal_gamma, type=float,
                           help='use focal loss with the given gamma')
        assert not cls.focal_detach
        group.add_argument('--focal-detach', default=False, action='store_true')
        assert cls.focal_clamp
        group.add_argument('--no-focal-clamp', dest='focal_clamp',
                           default=True, action='store_false')
        group.add_argument('--bce-min', default=cls.min_bce, type=float,
                           help='gradient clipped below')
        group.add_argument('--bce-soft-clamp', default=cls.soft_clamp_value, type=float,
                           help='soft clamp for BCE')
        group.add_argument('--bce-background-clamp', default=cls.background_clamp, type=float,
                           help='background clamp for BCE')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.background_weight = args.background_weight
        cls.focal_alpha = args.focal_alpha
        cls.focal_gamma = args.focal_gamma
        cls.focal_detach = args.focal_detach
        cls.focal_clamp = args.focal_clamp
        cls.min_bce = args.bce_min
        cls.soft_clamp_value = args.bce_soft_clamp
        cls.background_clamp = args.bce_background_clamp

    def forward(self, x, t):  # pylint: disable=arguments-differ
        t_zeroone = t.clone()
        t_zeroone[t_zeroone > 0.0] = 1.0
        # x = torch.clamp(x, -20.0, 20.0)
        if self.background_clamp is not None:
            bg_clamp_mask = (t_zeroone == 0.0) & (x < self.background_clamp)
            x[bg_clamp_mask] = self.background_clamp
        bce = torch.nn.functional.binary_cross_entropy_with_logits(
            x, t_zeroone, reduction='none')
        # torch.clamp_max_(bce, 10.0)
        if self.soft_clamp is not None:
            bce = self.soft_clamp(bce)
        if self.min_bce > 0.0:
            torch.clamp_min_(bce, self.min_bce)

        if self.focal_gamma != 0.0:
            p = torch.sigmoid(x)
            pt = p * t_zeroone + (1 - p) * (1 - t_zeroone)
            # Above code is more stable than deriving pt from bce: pt = torch.exp(-bce)

            if self.focal_clamp and self.min_bce > 0.0:
                pt_threshold = math.exp(-self.min_bce)
                torch.clamp_max_(pt, pt_threshold)

            focal = 1.0 - pt
            if self.focal_gamma != 1.0:
                focal = (focal + 1e-4)**self.focal_gamma

            if self.focal_detach:
                focal = focal.detach()

            bce = focal * bce

        if self.focal_alpha == 0.5:
            bce = 0.5 * bce
        elif self.focal_alpha >= 0.0:
            alphat = self.focal_alpha * t_zeroone + (1 - self.focal_alpha) * (1 - t_zeroone)
            bce = alphat * bce

        weight_mask = t_zeroone != t
        bce[weight_mask] = bce[weight_mask] * t[weight_mask]

        if self.background_weight != 1.0:
            bg_weight = torch.ones_like(t, requires_grad=False)
            bg_weight[t == 0] *= self.background_weight
            bce = bce * bg_weight

        return bce


class BceDistance(Bce):
    def forward(self, x, t):
        t_sign = t.clone()
        t_sign[t > 0.0] = 1.0
        t_sign[t <= 0.0] = -1.0

        # construct target location relative to x but without backpropagating through x
        x_detached = x.detach()
        focal_loss_modification = 1.0
        p_bar = 1.0 / (1.0 + torch.exp(t_sign * x_detached))
        if self.focal_alpha:
            focal_loss_modification *= self.focal_alpha
        if self.focal_gamma == 1.0:
            p = 1.0 - p_bar

            # includes simplifications for numerical stability
            # neg_ln_p = torch.log(1 + torch.exp(-t_sign * x_detached))
            # even more stability:
            neg_ln_p = torch.nn.functional.softplus(-t_sign * x_detached)

            focal_loss_modification = focal_loss_modification * (p_bar + p * neg_ln_p)
        elif self.focal_gamma > 0.0:
            p = 1.0 - p_bar
            neg_ln_p = torch.nn.functional.softplus(-t_sign * x_detached)

            focal_loss_modification = focal_loss_modification * (
                p_bar ** self.focal_gamma
                + self.focal_gamma * p_bar ** (self.focal_gamma - 1.0) * p * neg_ln_p
            )
        elif self.focal_gamma == 0.0:
            pass
        else:
            raise NotImplementedError
        target = x_detached + t_sign * p_bar * focal_loss_modification

        # construct distance with x that backpropagates gradients
        d = x - target

        # background clamp
        if self.background_clamp:
            d[(x_detached < self.background_clamp) & (t_sign == -1.0)] = 0.0

        return d


class BceL2(BceDistance):
    def forward(self, x, t):
        d = super().forward(x, t)
        return torch.nn.functional.smooth_l1_loss(d, torch.zeros_like(d), reduction='none')


class Scale(torch.nn.Module):
    b = 1.0
    log_space = False
    relative = True
    relative_eps = 0.1
    clip = None
    soft_clamp_value = 5.0

    def __init__(self):
        super().__init__()

        self.soft_clamp = None
        if self.soft_clamp_value:
            self.soft_clamp = SoftClamp(self.soft_clamp_value)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('Scale Loss')
        group.add_argument('--b-scale', default=cls.b, type=float,
                           help='Laplace width b for scale loss')
        assert not cls.log_space
        group.add_argument('--scale-log', default=False, action='store_true')
        group.add_argument('--scale-soft-clamp', default=cls.soft_clamp_value, type=float,
                           help='soft clamp for scale')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.b = args.b_scale
        cls.log_space = args.scale_log
        if args.scale_log:
            cls.relative = False
        cls.soft_clamp_value = args.scale_soft_clamp

    def forward(self, x, t):  # pylint: disable=arguments-differ
        assert not (self.log_space and self.relative)

        x = torch.nn.functional.softplus(x)
        d = torch.nn.functional.l1_loss(
            x if not self.log_space else torch.log(x),
            t if not self.log_space else torch.log(t),
            reduction='none',
        )
        if self.clip is not None:
            d = torch.clamp(d, self.clip[0], self.clip[1])

        denominator = self.b
        if self.relative:
            denominator = self.b * (self.relative_eps + t)
        d = d / denominator

        if self.soft_clamp is not None:
            d = self.soft_clamp(d)

        loss = torch.nn.functional.smooth_l1_loss(d, torch.zeros_like(d), reduction='none')
        return loss


class ScaleDistance(Scale):
    def forward(self, x, t):
        x = torch.nn.functional.softplus(x)
        d = 1.0 / self.b * (x - t)
        d[torch.isnan(d)] = 0.0
        return d


class Laplace(torch.nn.Module):
    """Loss based on Laplace Distribution.

    Loss for a single two-dimensional vector (x1, x2) with radial
    spread b and true (t1, t2) vector.
    """

    weight = None
    norm_clip = None
    soft_clamp_value = 5.0

    def __init__(self):
        super().__init__()

        self.soft_clamp = None
        if self.soft_clamp_value:
            self.soft_clamp = SoftClamp(self.soft_clamp_value)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('Laplace Loss')
        group.add_argument('--laplace-soft-clamp', default=cls.soft_clamp_value, type=float,
                           help='soft clamp for Laplace')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.soft_clamp_value = args.laplace_soft_clamp

    def forward(self, x1, x2, logb, t1, t2, bmin):
        # left derivative of sqrt at zero is not defined, so prefer torch.norm():
        # https://github.com/pytorch/pytorch/issues/2421
        # norm = torch.sqrt((x1 - t1)**2 + (x2 - t2)**2)
        # norm = (torch.stack((x1, x2)) - torch.stack((t1, t2))).norm(dim=0)
        # norm = (
        #     torch.nn.functional.l1_loss(x1, t1, reduction='none')
        #     + torch.nn.functional.l1_loss(x2, t2, reduction='none')
        # )
        # While torch.norm is a special treatment at zero, it does produce
        # large gradients for tiny values (as it should).
        # Similar to BatchNorm, we introduce a physically irrelevant epsilon
        # that stabilizes the gradients for small norms.
        # norm = torch.sqrt((x1 - t1)**2 + (x2 - t2)**2 + torch.clamp_min(bmin**2, 0.0001))
        # norm = torch.stack((x1 - t1, x2 - t2, torch.clamp_min(bmin, 0.01))).norm(dim=0)
        norm = torch.stack((x1 - t1, x2 - t2, bmin)).norm(dim=0)
        if self.norm_clip is not None:
            norm = torch.clamp(norm, self.norm_clip[0], self.norm_clip[1])

        # constrain range of logb
        # logb = logb + 3.0  # shift logb such that weight decay relaxes to large b
        # low range constraint: prevent strong confidence when overfitting
        # high range constraint: force some data dependence
        logb = 3.0 * torch.tanh(logb / 3.0)
        # b = torch.nn.functional.softplus(b)
        # b = torch.max(b, bmin)
        # b_plus_bmin = torch.nn.functional.softplus(b) + bmin
        # b_plus_bmin = 20.0 * torch.sigmoid(b / 20.0) + bmin
        # logb = -3.0 + torch.nn.functional.softplus(logb + 3.0)
        # log_bmin = torch.log(bmin)
        # logb = log_bmin + torch.nn.functional.softplus(logb - log_bmin)

        # ln(2) = 0.694
        # losses = torch.log(b_plus_bmin) + norm / b_plus_bmin
        scaled_norm = norm * torch.exp(-logb)
        if self.soft_clamp is not None:
            scaled_norm = self.soft_clamp(scaled_norm)
        losses = logb + scaled_norm
        if self.weight is not None:
            losses = losses * self.weight
        return losses


class RegressionDistance:
    @staticmethod
    def __call__(x_regs, t_regs, t_sigma_min, t_scales):
        d = x_regs - t_regs
        d = torch.cat([d, t_sigma_min], dim=2)

        # L2 distance for coordinate pair
        d_shape = d.shape
        d = d.reshape(d_shape[0], d_shape[1], -1, 3, d_shape[-2], d_shape[-1])
        d[torch.isnan(d)] = float('inf')
        d = torch.linalg.norm(d, ord=2, dim=3)
        d[~torch.isfinite(d)] = 0.0

        # 68% inside of t_sigma
        if t_scales.shape[2]:
            t_sigma = 0.5 * t_scales
            t_sigma[torch.isnan(t_sigma)] = 0.5  # assume a sigma when not given
        elif t_regs.shape[2] == 4:
            # two regressions without scales is detection, i.e. second
            # regression targets are width and height
            t_scales = torch.linalg.norm(0.5 * t_regs[:, :, 2:], ord=2, dim=2, keepdim=True)
            t_sigma = 0.5 * t_scales
            t_sigma[torch.isnan(t_sigma)] = 5.0  # assume a sigma when not given
            t_sigma = torch.repeat_interleave(t_sigma, 2, dim=2)
        else:
            t_sigma = 0.5
        d = 1.0 / t_sigma * d

        return d


class RegressionLoss:
    @staticmethod
    def __call__(x_regs, t_regs, t_sigma_min, t_scales):
        """Only t_regs is guaranteed to be valid.
        Imputes t_sigma_min and and t_scales with guesses.
        """
        x_regs = x_regs.reshape(t_regs.shape[0], t_regs.shape[1] // 2, 2)
        t_regs = t_regs.reshape(t_regs.shape[0], t_regs.shape[1] // 2, 2)
        t_sigma_min = t_sigma_min.unsqueeze(-1)
        t_scales = t_scales.unsqueeze(-1)

        d = x_regs - t_regs
        t_sigma_min_imputed = t_sigma_min.clone()
        t_sigma_min_imputed[torch.isnan(t_sigma_min)] = 0.1
        d = torch.cat([d, t_sigma_min_imputed], dim=2)

        # L2 distance for coordinate pair
        d = torch.linalg.norm(d, ord=2, dim=2, keepdim=True)

        # 68% inside of t_sigma
        if t_scales.shape[1]:
            t_sigma = 0.5 * t_scales
            t_sigma[torch.isnan(t_sigma)] = 0.5
        elif t_regs.shape[1] == 2:
            # Two regressions without scales is detection, i.e. second
            # regression targets are width and height.
            # Assume a regression sigma of 10% of the scale.
            t_scales = torch.linalg.norm(t_regs[:, 1:2], ord=2, dim=2, keepdim=True)
            t_sigma = 0.1 * t_scales
            t_sigma = torch.repeat_interleave(t_sigma, 2, dim=1)
        else:
            t_sigma = 0.5
        d = 1.0 / t_sigma * d

        return d


def l1_loss(x1, x2, _, t1, t2, weight=None):
    """L1 loss.

    Loss for a single two-dimensional vector (x1, x2)
    true (t1, t2) vector.
    """
    losses = torch.sqrt((x1 - t1)**2 + (x2 - t2)**2)
    if weight is not None:
        losses = losses * weight
    return losses


def logl1_loss(logx, t, **kwargs):
    """Swap in replacement for functional.l1_loss."""
    return torch.nn.functional.l1_loss(
        logx, torch.log(t), **kwargs)


class SmoothL1:
    r_smooth = 0.0

    def __init__(self, *, scale_required=True):
        self.scale = None
        self.scale_required = scale_required

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('Bce Loss')
        group.add_argument('--r-smooth', type=float, default=cls.r_smooth,
                           help='r_{smooth} for SmoothL1 regressions')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.r_smooth = args.r_smooth

    def __call__(self, x1, x2, _, t1, t2, weight=None):
        """L1 loss.

        Loss for a single two-dimensional vector (x1, x2)
        true (t1, t2) vector.
        """
        if self.scale_required and self.scale is None:
            raise Exception
        if self.scale is None:
            self.scale = 1.0

        r = self.r_smooth * self.scale
        d = torch.sqrt((x1 - t1)**2 + (x2 - t2)**2)
        smooth_regime = d < r

        smooth_loss = 0.5 / r[smooth_regime] * d[smooth_regime] ** 2
        linear_loss = d[smooth_regime == 0] - (0.5 * r[smooth_regime == 0])
        losses = torch.cat((smooth_loss, linear_loss))

        if weight is not None:
            losses = losses * weight

        self.scale = None
        return torch.sum(losses)



from collections import defaultdict
import logging
from typing import Optional

from .cifcaf import CifCaf, CifCafDense
from .cifdet import CifDet
from .decoder import Decoder
from .multi import Multi
from .pose_similarity import PoseSimilarity
from .track_base import TrackBase
from .tracking_pose import TrackingPose
from . import utils
from ..profiler import Profiler  # , TorchProfiler

LOG = logging.getLogger(__name__)

DECODERS = {CifDet, CifCaf, CifCafDense, PoseSimilarity, TrackingPose}


def cli(parser, *, workers=None):
    group = parser.add_argument_group('decoder configuration')

    available_decoders = [dec.__name__.lower() for dec in DECODERS]
    group.add_argument('--decoder', default=None, nargs='+',
                       help='Decoders to be considered: {}.'.format(available_decoders))

    assert utils.CifSeeds.get_threshold() == utils.CifDetSeeds.get_threshold()
    group.add_argument('--seed-threshold', default=utils.CifSeeds.get_threshold(), type=float,
                       help='minimum threshold for seeds')
    assert CifDet.instance_threshold == utils.nms.Keypoints.get_instance_threshold()
    group.add_argument('--instance-threshold', type=float, default=None,
                       help=('filter instances by score (default is 0.0 with '
                             '--force-complete-pose and {} otherwise)'
                             ''.format(utils.nms.Keypoints.get_instance_threshold())))
    group.add_argument('--decoder-workers', default=workers, type=int,
                       help='number of workers for pose decoding')

    group.add_argument('--profile-decoder', nargs='?', const='profile_decoder.prof', default=None,
                       help='specify out .prof file or nothing for default file name')

    group = parser.add_argument_group('CifCaf decoders')
    group.add_argument('--cif-th', default=utils.CifHr.get_threshold(), type=float,
                       help='cif threshold')
    group.add_argument('--caf-th', default=utils.CafScored.get_default_score_th(), type=float,
                       help='caf threshold')

    TrackBase.cli(parser)
    for dec in DECODERS:
        dec.cli(parser)


def configure(args):
    if args.instance_threshold is None:
        if args.force_complete_pose:
            args.instance_threshold = 0.0
        else:
            args.instance_threshold = utils.nms.Keypoints.get_instance_threshold()

    # configure Factory
    Factory.decoder_request_from_args(args.decoder)
    Factory.profile = args.profile_decoder

    # configure CifHr
    utils.CifHr.set_threshold(args.cif_th)

    # configure CifSeeds
    utils.CifSeeds.set_threshold(args.seed_threshold)
    utils.CifDetSeeds.set_threshold(args.seed_threshold)

    # configure CafScored
    utils.CafScored.set_default_score_th(args.caf_th)

    # configure generators
    Decoder.default_worker_pool = args.decoder_workers

    # configure instance threshold
    utils.nms.Keypoints.set_instance_threshold(args.instance_threshold)
    CifDet.instance_threshold = args.instance_threshold

    TrackBase.configure(args)
    for dec in DECODERS:
        dec.configure(args)


class Factory:
    decoder_request: Optional[dict] = None
    profile = False

    @classmethod
    def decoder_request_from_args(cls, list_str):
        if list_str is None:
            cls.decoder_request = None
            return

        cls.decoder_request = defaultdict(list)
        for dec_str in list_str:
            # pylint: disable=unsupported-assignment-operation,unsupported-membership-test,unsubscriptable-object
            if ':' not in dec_str:
                if dec_str not in cls.decoder_request:
                    cls.decoder_request[dec_str] = []
                continue

            dec_str, _, index = dec_str.partition(':')
            index = int(index)
            cls.decoder_request[dec_str].append(index)

        LOG.debug('setup decoder request: %s', cls.decoder_request)

    @classmethod
    def decoders(cls, head_metas):
        def per_class(request, dec_class):
            class_name = dec_class.__name__.lower()
            if request is not None \
               and class_name not in request:
                return []
            decoders = sorted(dec_class.factory(head_metas), key=lambda d: d.priority, reverse=True)
            for dec_i, dec in enumerate(decoders):
                dec.request_index = dec_i
            if request is not None:
                indices = set(request[class_name])
                decoders = (d for i, d in enumerate(decoders) if i in indices)
            return decoders

        decoders = [d for dec_class in DECODERS for d in per_class(cls.decoder_request, dec_class)]
        decoders = list(sorted(decoders, key=lambda d: d.priority, reverse=True))
        LOG.debug('created %d decoders', len(decoders))

        if not decoders:
            LOG.warning('no decoders found for heads %s', [meta.name for meta in head_metas])
        elif len(decoders) == 1:
            pass
        elif cls.decoder_request is None:
            LOG.info(
                'No specific decoder requested. Using the first one from:\n'
                '%s\n'
                'Use any of the above arguments to select one or multiple '
                'decoders and to suppress this message.',
                '\n'.join(
                    f'  --decoder={dec.__class__.__name__.lower()}:{dec.request_index}'
                    for dec in decoders
                )
            )
            decoders = [decoders[0]]

        return decoders

    @classmethod
    def __call__(cls, head_metas):
        """Instantiate decoders."""
        LOG.debug('head names = %s', [meta.name for meta in head_metas])
        decoders = cls.decoders(head_metas)

        if cls.profile:
            decode = decoders[0]
            decode.__class__.__call__ = Profiler(
                decode.__call__, out_name=cls.profile)
            # decode.__class__.__call__ = TorchProfiler(
            #     decode.__call__, out_name=cls.profile)

        return Multi(decoders)

        # TODO implement!
        # skeleton = head_nets[1].meta.skeleton
        # if dense_connections:
        #     field_config.confidence_scales = (
        #         [1.0 for _ in skeleton] +
        #         [dense_coupling for _ in head_nets[2].meta.skeleton]
        #     )
        #     skeleton = skeleton + head_nets[2].meta.skeleton


factory = Factory.__call__



import torch

import openpifpaf


class Cifar10Net(openpifpaf.network.BaseNetwork):
    """Small network for Cifar10."""
    def __init__(self):
        super().__init__('cifar10net', stride=16, out_features=128)
        self.conv1 = torch.nn.Conv2d(3, 16, 3, 2, 1)
        self.conv2 = torch.nn.Conv2d(16, 32, 3, 2, 1)
        self.conv3 = torch.nn.Conv2d(32, 64, 3, 2, 1)
        self.conv4 = torch.nn.Conv2d(64, 128, 3, 2, 1)

    def forward(self, x):
        x = torch.nn.functional.relu(self.conv1(x))
        x = torch.nn.functional.relu(self.conv2(x))
        x = torch.nn.functional.relu(self.conv3(x))
        x = torch.nn.functional.relu(self.conv4(x))
        return x



"""CLI to convert a single-image checkpoint to a tracking checkpoint."""

import argparse
import logging
import os

import torch

import openpifpaf

LOG = logging.getLogger(__name__)


def cli():
    parser = argparse.ArgumentParser(description=__doc__)

    openpifpaf.plugin.register()
    openpifpaf.logger.cli(parser)
    openpifpaf.network.Factory.cli(parser)

    parser.add_argument('-o', '--output', default=None)
    args = parser.parse_args()

    openpifpaf.logger.configure(args, LOG)
    openpifpaf.network.Factory.configure(args)

    assert args.checkpoint, 'have to specify a checkpoint as input'
    if args.output is None:
        basename = 't' + os.path.basename(args.checkpoint)
        if not basename.endswith('.pkl'):
            basename += '-{}.pkl'.format(openpifpaf.__version__)
        args.output = os.path.join('outputs', basename)

    return args


def main():
    args = cli()
    model, _ = openpifpaf.network.Factory().factory()
    model.base_net = openpifpaf.network.TrackingBase(model.base_net)
    # model = openpifpaf.network.batchrenorm.BatchRenorm2d.convert_to(model)

    LOG.info('saving %s', args.output)
    torch.save({
        'model': model,
        'epoch': 0,
        'meta': {
            'image-source': args.checkpoint,
        },
    }, args.output)


if __name__ == '__main__':
    main()



from abc import ABCMeta, abstractmethod


class Preprocess(metaclass=ABCMeta):
    """Preprocess an image with annotations and meta information."""
    @abstractmethod
    def __call__(self, image, anns, meta):
        """Implementation of preprocess operation."""



import copy
import logging
import math

import numpy as np
import PIL
import torch

from .pad import CenterPad
from .preprocess import Preprocess
from .. import utils

try:
    import scipy
except ImportError:
    scipy = None

LOG = logging.getLogger(__name__)


def rotate(image, anns, meta, angle):
    meta = copy.deepcopy(meta)
    anns = copy.deepcopy(anns)

    LOG.debug('rotation angle = %f', angle)
    w, h = image.size
    assert meta['rotation']['angle'] == 0.0
    meta['rotation']['angle'] = angle
    meta['rotation']['width'] = w
    meta['rotation']['height'] = h

    # rotate image
    if angle != 0.0:
        im_np = np.asarray(image)
        if im_np.shape[0] == im_np.shape[1] and angle == 90:
            im_np = np.swapaxes(im_np, 0, 1)
            im_np = np.flip(im_np, axis=0)
        elif im_np.shape[0] == im_np.shape[1] and angle == 270:
            im_np = np.swapaxes(im_np, 0, 1)
            im_np = np.flip(im_np, axis=1)
        elif im_np.shape[0] == im_np.shape[1] and angle == 180:
            im_np = np.flip(im_np, axis=0)
            im_np = np.flip(im_np, axis=1)
        else:
            fill_value = int(torch.randint(0, 255, (1,)).item())
            im_np = scipy.ndimage.rotate(im_np, angle=angle, cval=fill_value, reshape=False)
        image = PIL.Image.fromarray(im_np)
    LOG.debug('rotated by = %f degrees', angle)

    # rotate keypoints
    cangle = math.cos(angle / 180.0 * math.pi)
    sangle = math.sin(angle / 180.0 * math.pi)
    for ann in anns:
        xy = ann['keypoints'][:, :2]
        x_old = xy[:, 0].copy() - (w - 1) / 2
        y_old = xy[:, 1].copy() - (h - 1) / 2
        xy[:, 0] = (w - 1) / 2 + cangle * x_old + sangle * y_old
        xy[:, 1] = (h - 1) / 2 - sangle * x_old + cangle * y_old
        ann['bbox'] = utils.rotate_box(ann['bbox'], w - 1, h - 1, angle)

    LOG.debug('meta before: %s', meta)
    meta['valid_area'] = utils.rotate_box(meta['valid_area'], w - 1, h - 1, angle)
    # fix valid area to be inside original image dimensions
    original_valid_area = meta['valid_area'].copy()
    meta['valid_area'][0] = np.clip(meta['valid_area'][0], 0, w - 1)
    meta['valid_area'][1] = np.clip(meta['valid_area'][1], 0, h - 1)
    new_rb_corner = original_valid_area[:2] + original_valid_area[2:]
    new_rb_corner[0] = np.clip(new_rb_corner[0], 0, w - 1)
    new_rb_corner[1] = np.clip(new_rb_corner[1], 0, h - 1)
    meta['valid_area'][2:] = new_rb_corner - meta['valid_area'][:2]
    LOG.debug('meta after: %s', meta)

    return image, anns, meta


def _prepad(image, anns, meta, angle):
    if abs(angle) < 0.3:
        return image, anns, meta

    w, h = image.size
    cos_angle = math.cos(abs(angle) * math.pi / 180.0)
    sin_angle = math.sin(abs(angle) * math.pi / 180.0)
    LOG.debug('angle = %f, cos = %f, sin = %f', angle, cos_angle, sin_angle)
    padded_size = (
        int(w * cos_angle + h * sin_angle) + 1,
        int(h * cos_angle + w * sin_angle) + 1,
    )
    center_pad = CenterPad(padded_size)
    return center_pad(image, anns, meta)


class RotateBy90(Preprocess):
    """Randomly rotate by multiples of 90 degrees."""

    def __init__(self, angle_perturbation=0.0, fixed_angle=None, prepad=False):
        super().__init__()

        self.angle_perturbation = angle_perturbation
        self.fixed_angle = fixed_angle
        self.prepad = prepad

    def __call__(self, image, anns, meta):
        if self.fixed_angle is not None:
            angle = self.fixed_angle
        else:
            rnd1 = float(torch.rand(1).item())
            angle = int(rnd1 * 4.0) * 90.0
            sym_rnd2 = (float(torch.rand(1).item()) - 0.5) * 2.0
            angle += sym_rnd2 * self.angle_perturbation

        if self.prepad:
            image, anns, meta = _prepad(image, anns, meta, angle)
        return rotate(image, anns, meta, angle)


class RotateUniform(Preprocess):
    """Rotate by a random angle uniformly drawn from a given angle range."""

    def __init__(self, max_angle=30.0, prepad=True):
        super().__init__()
        self.max_angle = max_angle
        self.prepad = prepad

    def __call__(self, image, anns, meta):
        sym_rnd = (float(torch.rand(1).item()) - 0.5) * 2.0
        angle = sym_rnd * self.max_angle

        if self.prepad:
            image, anns, meta = _prepad(image, anns, meta, angle)
        return rotate(image, anns, meta, angle)



from collections import defaultdict
import logging

LOG = logging.getLogger(__name__)


class Signal:
    """In-process Signal infrastructure.

    Objects can subscribe to 'events'. Events are triggered with the emit()
    function and are fanned out to all subscribers.
    """

    subscribers = defaultdict(list)

    @classmethod
    def emit(cls, name, *args, **kwargs):
        subscribers = cls.subscribers.get(name, [])
        LOG.debug('emit %s to %d subscribers', name, len(subscribers))
        for subscriber in subscribers:
            subscriber(*args, **kwargs)

    @classmethod
    def subscribe(cls, name, subscriber):
        LOG.debug('subscribe to %s', name)
        cls.subscribers[name].append(subscriber)



import argparse
import logging
import time

import numpy as np
import torch

try:
    import cv2  # pylint: disable=import-error
except ImportError:
    cv2 = None

import PIL
try:
    import PIL.ImageGrab
except ImportError:
    pass

try:
    import mss
except ImportError:
    mss = None

LOG = logging.getLogger(__name__)


# pylint: disable=abstract-method
class Stream(torch.utils.data.IterableDataset):
    horizontal_flip = None
    rotate = None
    crop = None
    scale = 1.0
    start_frame = None
    start_msec = None
    max_frames = None

    def __init__(self, source, *,
                 preprocess=None,
                 with_raw_image=True):
        super().__init__()

        self.source = source
        self.preprocess = preprocess
        self.with_raw_image = with_raw_image

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        """Command line interface (CLI) to extend argument parser."""
        group = parser.add_argument_group('Stream')
        group.add_argument('--horizontal-flip', default=False, action='store_true',
                           help='mirror input image')
        group.add_argument('--scale', default=1.0, type=float,
                           help='input image scale factor')
        group.add_argument('--start-frame', type=int, default=None,
                           help='start frame')
        group.add_argument('--start-msec', type=float, default=None,
                           help='start millisecond')
        group.add_argument('--crop', type=int, nargs=4, default=None,
                           help='left top right bottom')
        group.add_argument('--rotate', default=None, choices=('left', 'right', '180'),
                           help='rotate')
        group.add_argument('--max-frames', type=int, default=None,
                           help='max frames')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Take the parsed argument parser output and configure class variables."""
        cls.horizontal_flip = args.horizontal_flip
        cls.scale = args.scale
        cls.start_frame = args.start_frame
        cls.start_msec = args.start_msec
        cls.crop = args.crop
        cls.rotate = args.rotate
        cls.max_frames = args.max_frames

    # pylint: disable=unsubscriptable-object
    def preprocessing(self, image):
        if self.scale != 1.0:
            image = cv2.resize(image, None, fx=self.scale, fy=self.scale)
            LOG.debug('resized image size: %s', image.shape)
        if self.horizontal_flip:
            image = image[:, ::-1]
        if self.crop:
            if self.crop[0]:
                image = image[:, self.crop[0]:]
            if self.crop[1]:
                image = image[self.crop[1]:, :]
            if self.crop[2]:
                image = image[:, :-self.crop[2]]
            if self.crop[3]:
                image = image[:-self.crop[3], :]
        if self.rotate == 'left':
            image = np.swapaxes(image, 0, 1)
            image = np.flip(image, axis=0)
        elif self.rotate == 'right':
            image = np.swapaxes(image, 0, 1)
            image = np.flip(image, axis=1)
        elif self.rotate == '180':
            image = np.flip(image, axis=0)
            image = np.flip(image, axis=1)

        image_pil = PIL.Image.fromarray(np.ascontiguousarray(image))
        meta = {
            'hflip': False,
            'offset': np.array([0.0, 0.0]),
            'scale': np.array([1.0, 1.0]),
            'valid_area': np.array([0.0, 0.0, image_pil.size[0], image_pil.size[1]]),
        }
        processed_image, anns, meta = self.preprocess(image_pil, [], meta)
        return image, processed_image, anns, meta

    # pylint: disable=too-many-branches
    def __iter__(self):
        if self.source == 'screen':
            capture = 'screen'
            if mss is None:
                print('!!!!!!!!!!! install mss (pip install mss) for faster screen grabs')
        else:
            capture = cv2.VideoCapture(self.source)
            if self.start_frame:
                capture.set(cv2.CAP_PROP_POS_FRAMES, self.start_frame)
            if self.start_msec:
                capture.set(cv2.CAP_PROP_POS_MSEC, self.start_msec)

        frame_start = 0 if not self.start_frame else self.start_frame
        frame_i = frame_start
        while True:
            frame_i += 1
            if self.max_frames and frame_i - frame_start > self.max_frames:
                LOG.info('reached max frames %d', self.max_frames)
                break

            if capture == 'screen':
                if mss is None:
                    image = np.asarray(PIL.ImageGrab.grab().convert('RGB'))
                else:
                    with mss.mss() as sct:
                        monitor = sct.monitors[1]
                        image = np.asarray(sct.grab(monitor))[:, :, 2::-1]
            else:
                _, image = capture.read()
                if image is not None:
                    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

            if image is None:
                LOG.info('no more images captured')
                break

            start_preprocess = time.perf_counter()
            image, processed_image, anns, meta = self.preprocessing(image)
            meta['frame_i'] = frame_i
            meta['preprocessing_s'] = time.perf_counter() - start_preprocess

            if self.with_raw_image:
                yield image, processed_image, anns, meta
            else:
                yield processed_image, anns, meta



import copy
import logging

import numpy as np

import openpifpaf

LOG = logging.getLogger(__name__)


class NormalizePosetrack(openpifpaf.transforms.Preprocess):
    def __init__(self, ignore_missing_bbox=False, fix_annotations=True):
        self.ignore_missing_bbox = ignore_missing_bbox
        self.fix_annotations = fix_annotations

    @staticmethod
    def add_crowd_annotations(anns, image_info):
        ignore_regions = []
        if 'ignore_regions_x' in image_info:
            ignore_regions = list(
                zip(image_info['ignore_regions_x'],
                    image_info['ignore_regions_y'])
            )

        # add ignore regions to annotations
        anns += [
            {
                'bbox': [min(xs), min(ys), max(xs) - min(xs), max(ys) - min(ys)],
                'keypoints': [(x, y, 0.0) for x, y in zip(xs, ys)],
                'iscrowd': True,
                'track_id': -1,
            }
            for xs, ys in ignore_regions if xs and ys
        ]

        return anns

    def normalize_annotations(self, anns, valid_area, image_id):
        # convert as much data as possible to numpy arrays to avoid every float
        # being turned into its own torch.Tensor()
        for ann in anns:
            ann['image_id'] = image_id
            ann['keypoints'] = np.asarray(ann['keypoints'], dtype=np.float32).reshape(-1, 3)

            # Posetrack does not distinguish between visible keypoints and
            # invisible keypoints. Treat all annotated keypoints as visible.
            ann['keypoints'][ann['keypoints'][:, 2] > 0.0, 2] = 2.0

            # Fix keypoints.
            # PoseTrack data contains some bad data.
            if self.fix_annotations:
                ann['keypoints'][ann['keypoints'][:, 0] < valid_area[0], 2] = 0.0
                ann['keypoints'][ann['keypoints'][:, 1] < valid_area[1], 2] = 0.0
                ann['keypoints'][ann['keypoints'][:, 0] > valid_area[0] + valid_area[2], 2] = 0.0
                ann['keypoints'][ann['keypoints'][:, 1] > valid_area[1] + valid_area[3], 2] = 0.0

            if 'bbox' in ann:
                ann['bbox'] = np.asarray(ann['bbox'], dtype=np.float32)
                # v = ann['keypoints'][:, 2] > 0
                # print(ann['bbox'], min(ann['keypoints'][v, 0]), min(ann['keypoints'][v, 1]))
                # if not np.all(ann['keypoints'][v, 0] >= ann['bbox'][0]) or \
                #    not np.all(ann['keypoints'][v, 1] >= ann['bbox'][1]) or \
                #    not np.all(ann['keypoints'][v, 0] <= ann['bbox'][0] + ann['bbox'][2]) or \
                #    not np.all(ann['keypoints'][v, 1] <= ann['bbox'][1] + ann['bbox'][3]):
                #     print(ann)
                #     assert False
            else:
                ann['bbox'] = np.zeros((4,), dtype=np.float32)
                if not self.ignore_missing_bbox:
                    assert all(c == 0.0 for c in ann['keypoints'][:, 2])

            if 'bbox_head' in ann:
                ann['bbox_head'] = np.asarray(ann['bbox_head'], dtype=np.float32)

            if 'iscrowd' not in ann:
                ann['iscrowd'] = False
                assert len(ann['keypoints']) == 17

            if not ann['iscrowd']:
                # no ear annotations
                assert ann['keypoints'][3, 2] == 0.0
                assert ann['keypoints'][4, 2] == 0.0

        return anns

    def __call__(self, image, anns, meta=None):
        meta = copy.deepcopy(meta)

        w, h = image.size
        meta_init = {
            'offset': np.array((0.0, 0.0)),
            'scale': np.array((1.0, 1.0)),
            'rotation': {'angle': 0.0, 'width': None, 'height': None},
            'valid_area': np.array((0.0, 0.0, w - 1, h - 1)),
            'hflip': False,
            'width_height': np.array((w, h)),
        }
        if meta is None:
            meta = meta_init
        else:
            for k, v in meta_init.items():
                if k in meta:
                    continue
                meta[k] = v

        image_info = anns['image']
        anns = copy.deepcopy(anns['annotations'])

        anns = self.add_crowd_annotations(anns, image_info)
        anns = self.normalize_annotations(anns, meta['valid_area'], image_info['frame_id'])
        return image, anns, meta


class NormalizeMOT(openpifpaf.transforms.Preprocess):
    @staticmethod
    def normalize_annotations(anns):
        anns = copy.deepcopy(anns)

        # convert as much data as possible to numpy arrays to avoid every float
        # being turned into its own torch.Tensor()
        for ann in anns:
            ann['keypoints'] = np.asarray(ann['keypoints'], dtype=np.float32).reshape(-1, 3)
            ann['bbox'] = np.asarray(ann['bbox'], dtype=np.float32)
            del ann['segmentation']

        return anns

    def __call__(self, image, anns, meta=None):
        anns = self.normalize_annotations(anns)

        if meta is None:
            w, h = image.size
            meta = {
                'offset': np.array((0.0, 0.0)),
                'scale': np.array((1.0, 1.0)),
                'valid_area': np.array((0.0, 0.0, w, h)),
                'hflip': False,
                'width_height': np.array((w, h)),
            }

        return image, anns, meta




from collections import defaultdict


def skeleton_mapping(kps_mapping):
    """Map the subset of keypoints from 0 to n-1"""
    map_sk = defaultdict(lambda: 100)  # map to 100 the keypoints not used
    for i, j in zip(kps_mapping, range(len(kps_mapping))):
        map_sk[i] = j
    return map_sk


def transform_skeleton(skeleton_orig, kps_mapping):
    """
    Transform the original apollo skeleton of 66 joints into a skeleton from 1 to n
    """
    map_sk = skeleton_mapping(kps_mapping)
    # skeleton = [[dic_sk[i], dic_sk[j]] for i, j in SKELETON]  # TODO
    skeleton = []
    for i, j in skeleton_orig:
        skeleton.append([map_sk[i] + 1, map_sk[j] + 1])   # skeleton starts from 1
    return skeleton



from . import tracking_heads
from .nets import model_defaults
from .tracking_base import TrackingBase
from ..signal import Signal

MODEL_MIGRATION = set()


# pylint: disable=protected-access
def model_migration(net_cpu):
    model_defaults(net_cpu)

    if not hasattr(net_cpu, 'process_heads'):
        net_cpu.process_heads = None

    for m in net_cpu.modules():
        if not hasattr(m, '_non_persistent_buffers_set'):
            m._non_persistent_buffers_set = set()

    for m in net_cpu.modules():
        if m.__class__.__name__ != 'InvertedResidualK':
            continue
        if not hasattr(m, 'branch1'):
            m.branch1 = None

    if not hasattr(net_cpu, 'head_nets') and hasattr(net_cpu, '_head_nets'):
        net_cpu.head_nets = net_cpu._head_nets

    for hn_i, hn in enumerate(net_cpu.head_nets):
        if not hn.meta.base_stride:
            hn.meta.base_stride = net_cpu.base_net.stride
        if hn.meta.head_index is None:
            hn.meta.head_index = hn_i
        if hn.meta.name == 'cif' and 'score_weights' not in vars(hn.meta):
            hn.meta.score_weights = [3.0] * 3 + [1.0] * (hn.meta.n_fields - 3)

    for mm in MODEL_MIGRATION:
        mm(net_cpu)


def fix_feature_cache(model):
    for m in model.modules():
        if not isinstance(m, TrackingBase):
            continue
        m.reset()


def subscribe_cache_reset(model):
    for m in model.modules():
        if not isinstance(m, TrackingBase):
            continue
        Signal.subscribe('eval_reset', m.reset)


def tcaf_shared_preprocessing(model):
    for m in model.modules():
        if not isinstance(m, tracking_heads.Tcaf):
            continue

        # pylint: disable=protected-access
        tracking_heads.Tcaf._global_feature_reduction = m.feature_reduction
        tracking_heads.Tcaf._global_feature_compute = m.feature_compute
        return


MODEL_MIGRATION.add(fix_feature_cache)
MODEL_MIGRATION.add(subscribe_cache_reset)
MODEL_MIGRATION.add(tcaf_shared_preprocessing)



"""Drawing primitives."""

from .animation_frame import AnimationFrame
from .canvas import Canvas, canvas, image_canvas, white_screen
from .cli import cli, configure
from .fields import arrows, boxes, boxes_wh, circles, margins, quiver
from .painters import CrowdPainter, DetectionPainter, KeypointPainter
from .annotation_painter import AnnotationPainter

__all__ = [
    'AnimationFrame',
    'Canvas',
    'arrows', 'boxes', 'boxes_wh', 'circles', 'margins', 'quiver',
    'CrowdPainter', 'DetectionPainter', 'KeypointPainter',
    'AnnotationPainter',
]



from .module import DataModule
from .multiloader import MultiLoader
from .multimodule import MultiDataModule

DATAMODULES = {}


def factory(dataset):
    if '-' in dataset:
        datamodules = [factory(ds) for ds in dataset.split('-')]
        return MultiDataModule(datamodules)

    if dataset not in DATAMODULES:
        raise Exception('dataset {} unknown'.format(dataset))
    return DATAMODULES[dataset]()


def cli(parser):
    group = parser.add_argument_group('generic data module parameters')
    group.add_argument('--dataset')
    group.add_argument('--loader-workers',
                       default=None, type=int,
                       help='number of workers for data loading')
    group.add_argument('--batch-size',
                       default=DataModule.batch_size, type=int,
                       help='batch size')
    group.add_argument('--dataset-weights', default=None, nargs='+', type=float,
                       help='n-1 weights for the datasets')

    for dm in DATAMODULES.values():
        dm.cli(parser)


def configure(args):
    DataModule.set_loader_workers(args.loader_workers if not args.debug else 0)
    DataModule.batch_size = args.batch_size
    MultiLoader.weights = args.dataset_weights

    for dm in DATAMODULES.values():
        dm.configure(args)



import PIL
import torch


from .. import transforms


class ImageList(torch.utils.data.Dataset):
    def __init__(self, image_paths, preprocess=None, with_raw_image=False):
        super().__init__()
        self.image_paths = image_paths
        self.preprocess = preprocess or transforms.EVAL_TRANSFORM
        self.with_raw_image = with_raw_image

    def __getitem__(self, index):
        image_path = self.image_paths[index]
        with open(image_path, 'rb') as f:
            image = PIL.Image.open(f).convert('RGB')

        anns = []
        meta = {
            'dataset_index': index,
            'file_name': image_path,
        }
        processed_image, anns, meta = self.preprocess(image, anns, meta)
        if self.with_raw_image:
            return image, processed_image, anns, meta
        return processed_image, anns, meta

    def __len__(self):
        return len(self.image_paths)


class PilImageList(torch.utils.data.Dataset):
    def __init__(self, images, preprocess=None, with_raw_image=False):
        super().__init__()
        self.images = images
        self.preprocess = preprocess or transforms.EVAL_TRANSFORM
        self.with_raw_image = with_raw_image

    def __getitem__(self, index):
        image = self.images[index].copy().convert('RGB')

        anns = []
        meta = {
            'dataset_index': index,
        }
        processed_image, anns, meta = self.preprocess(image, anns, meta)
        if self.with_raw_image:
            return image, processed_image, anns, meta
        return processed_image, anns, meta

    def __len__(self):
        return len(self.images)


class NumpyImageList(torch.utils.data.Dataset):
    def __init__(self, images, preprocess=None, with_raw_image=False):
        super().__init__()
        self.images = images
        self.preprocess = preprocess or transforms.EVAL_TRANSFORM
        self.with_raw_image = with_raw_image

    def __getitem__(self, index):
        image = PIL.Image.fromarray(self.images[index]).copy()

        anns = []
        meta = {
            'dataset_index': index,
        }
        processed_image, anns, meta = self.preprocess(image, anns, meta)
        if self.with_raw_image:
            return image, processed_image, anns, meta
        return processed_image, anns, meta

    def __len__(self):
        return len(self.images)



"""Benchmark."""

import argparse
from collections import namedtuple
from dataclasses import dataclass
import datetime
import json
import logging
import os
import subprocess
from typing import List

import pysparkling
try:
    import tabulate
except ImportError:
    tabulate = None

from . import __version__

LOG = logging.getLogger(__name__)


DEFAULT_CHECKPOINTS = [
    'resnet50',
    'shufflenetv2k16',
    'shufflenetv2k30',
]


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def cli():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.benchmark',
        description=__doc__,
        formatter_class=CustomFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=__version__))

    parser.add_argument('--output', default=None,
                        help='output file name')
    parser.add_argument('--checkpoints', default=DEFAULT_CHECKPOINTS, nargs='+',
                        help='checkpoints to evaluate')
    parser.add_argument('--iccv2019-ablation', default=False, action='store_true')
    parser.add_argument('--v012-ablation-1', default=False, action='store_true')
    parser.add_argument('--v012-ablation-2', default=False, action='store_true')
    parser.add_argument('--v012-ablation-3', default=False, action='store_true')
    parser.add_argument('--v012-ablation-4', default=False, action='store_true')
    group = parser.add_argument_group('logging')
    group.add_argument('--debug', default=False, action='store_true',
                       help='print debug messages')
    args, eval_args = parser.parse_known_args()

    logging.basicConfig(level=logging.INFO if not args.debug else logging.DEBUG)

    # default eval_args
    if not eval_args:
        eval_args = ['--loader-workers=8']

    # default loader workers
    if not any(l.startswith('--loader-workers') for l in eval_args):
        LOG.info('adding "--loader-workers=8" to the argument list')
        eval_args.append('--loader-workers=8')

    # default dataset
    if not any(l.startswith('--dataset') for l in eval_args):
        LOG.info('adding "--dataset=cocokp" to the argument list')
        eval_args.append('--dataset=cocokp')
        if not any(l.startswith('--coco-no-eval-annotation-filter') for l in eval_args):
            LOG.info('adding "--coco-no-eval-annotation-filter" to the argument list')
            eval_args.append('--coco-no-eval-annotation-filter')
        if not any(l.startswith('--force-complete-pose') for l in eval_args):
            LOG.info('adding "--force-complete-pose" to the argument list')
            eval_args.append('--force-complete-pose')
        if not any(l.startswith('--seed-threshold') for l in eval_args):
            LOG.info('adding "--seed-threshold=0.2" to the argument list')
            eval_args.append('--seed-threshold=0.2')
        if not any(l.startswith('--decoder') for l in eval_args):
            LOG.info('adding "--decoder=cifcaf:0" to the argument list')
            eval_args.append('--decoder=cifcaf:0')

    # generate a default output filename
    if args.output is None:
        now = datetime.datetime.now().strftime('%y%m%d-%H%M%S')
        args.output = 'outputs/benchmark-{}/'.format(now)
        os.makedirs(args.output)

    return args, eval_args


@dataclass
class Config:
    checkpoint: str
    suffix: str
    args: List[str]

    @property
    def name(self):
        return (self.checkpoint + self.suffix).replace('/', '-')


class Benchmark:
    def __init__(self, configs, output_folder, *,
                 reference_config=None, stat_filter=None, stat_scale=1.0):
        assert tabulate is not None, 'need tabulate: pip install tabulate'
        self.configs = configs
        self.output_folder = output_folder
        self.reference_config = reference_config
        self.stat_filter = stat_filter
        self.stat_scale = stat_scale

    def run(self):
        for config in self.configs:
            self.run_config(config)
        return self

    def run_config(self, config: Config):
        out_file = os.path.join(self.output_folder, config.name)
        if os.path.exists(out_file + '.stats.json'):
            LOG.warning('Output file %s exists already. Skipping.',
                        out_file + '.stats.json')
            return

        LOG.debug('Launching eval for %s.', config.name)
        cmd = [
            'python', '-m', 'openpifpaf.eval',
            '--output', out_file,
            '--checkpoint', config.checkpoint,
        ] + config.args
        LOG.info('eval command: %s', ' '.join(cmd))
        subprocess.run(cmd, check=True)

        # print intermediate output
        self.print_md()

    def stats(self):
        sc = pysparkling.Context()
        stats = (
            sc
            .wholeTextFiles(self.output_folder + '*.stats.json')
            .mapValues(json.loads)
            .map(lambda d: (d[0].replace('.stats.json', '').replace(self.output_folder, ''), d[1]))
            .collectAsMap()
        )
        LOG.debug('all data: %s', stats)
        return stats

    def stat_values(self, stat):
        return [
            v * self.stat_scale
            for l, v in zip(stat['text_labels'], stat['stats'])
            if self.stat_filter is None or l in self.stat_filter
        ]

    def stat_text_labels(self, stat):
        return [
            l
            for l in stat['text_labels']
            if self.stat_filter is None or l in self.stat_filter
        ]

    def print_md(self):
        """Pretty printing markdown"""
        stats = self.stats()
        first_stats = list(stats.values())[0]

        headers = ['Name']
        headers += self.stat_text_labels(first_stats)
        headers += ['t_total', 't_NN', 't_dec', 'size']

        reference_values = None
        if self.reference_config is not None:
            reference = stats.get(self.reference_config.name)
            if reference is not None:
                reference_values = self.stat_values(reference)

        table = []
        for name, data in sorted(stats.items(), key=lambda b_d: self.stat_values(b_d[1])[0]):
            values = self.stat_values(data)
            t = 1000.0 * data['total_time'] / data['n_images']
            tnn = 1000.0 * data['nn_time'] / data['n_images']
            tdec = 1000.0 * data['decoder_time'] / data['n_images']
            file_size = data['file_size'] / 1024 / 1024

            row = [f'[{name}]']

            if reference_values is not None and self.reference_config.name != name:
                values = [v - r for v, r in zip(values, reference_values)]
                row += [f'__{values[0]:+.1f}__']
                if len(values) > 1:
                    row += [f'{v:+.1f}' for v in values[1:]]

                t -= 1000.0 * reference['total_time'] / reference['n_images']
                tdec -= 1000.0 * reference['decoder_time'] / reference['n_images']
                tnn -= 1000.0 * reference['nn_time'] / reference['n_images']
                file_size -= reference['file_size'] / 1024 / 1024
                row += [f'{t:+.0f}ms', f'{tnn:+.0f}ms', f'{tdec:+.0f}ms', f'{file_size:+.1f}MB']
            else:
                row += [f'__{values[0]:.1f}__']
                if len(values) > 1:
                    row += [f'{v:.1f}' for v in values[1:]]
                row += [f'{t:.0f}ms', f'{tnn:.0f}ms', f'{tdec:.0f}ms', f'{file_size:.1f}MB']

            table.append(row)

        colalign = ['left'] + ['right'] * (len(headers) - 1)
        print(tabulate.tabulate(table, headers=headers, tablefmt='pipe', colalign=colalign))
        return self


def main():
    args, eval_args = cli()
    Ablation = namedtuple('Ablation', ['suffix', 'args'])
    ablations = [Ablation('', eval_args)]

    if args.iccv2019_ablation:
        ablations += [
            Ablation('.singlescale-max', eval_args + ['--connection-method=max']),
            Ablation('.singlescale', eval_args + ['--connection-method=blend']),
            Ablation('.multiscale-nohflip', eval_args + ['--connection-method=blend',
                                                         '--long-edge=961',
                                                         '--multi-scale',
                                                         '--no-multi-scale-hflip']),
            Ablation('.multiscale', eval_args + ['--connection-method=blend',
                                                 '--long-edge=961',
                                                 '--multi-scale']),
        ]
    if args.v012_ablation_1:
        ablations += [
            Ablation('.greedy', eval_args + ['--greedy']),
            Ablation('.no-reverse', eval_args + ['--no-reverse-match']),
            Ablation('.greedy.no-reverse', eval_args + ['--greedy', '--no-reverse-match']),
            Ablation('.greedy.dense', eval_args + [
                '--greedy', '--cocokp-with-dense', '--dense-connections']),
            Ablation('.dense', eval_args + [
                '--cocokp-with-dense', '--dense-connections']),
            Ablation('.dense.hierarchy', eval_args + [
                '--cocokp-with-dense', '--dense-connections=0.1']),
        ]
    if args.v012_ablation_2:
        ablations += [
            Ablation('.cifnr', eval_args + ['--ablation-cifseeds-no-rescore']),
            Ablation('.cifnr.nms', eval_args + ['--ablation-cifseeds-no-rescore',
                                                '--ablation-cifseeds-nms']),
            Ablation('.cafnr', eval_args + ['--ablation-caf-no-rescore']),
            Ablation('.nr.nms', eval_args + ['--ablation-cifseeds-no-rescore',
                                             '--ablation-cifseeds-nms',
                                             '--ablation-caf-no-rescore']),
        ]
    if args.v012_ablation_3:
        eval_args_nofc = [a for a in eval_args if not a.startswith('--force-complete')]
        ablations += [
            Ablation('.nofc', eval_args_nofc),
            Ablation('.nr.nms.nofc', eval_args_nofc + ['--ablation-cifseeds-no-rescore',
                                                       '--ablation-cifseeds-nms',
                                                       '--ablation-caf-no-rescore']),
        ]
    if args.v012_ablation_4:
        ablations += [
            Ablation('.indkp', eval_args + ['--ablation-independent-kp',
                                            '--keypoint-threshold=0.2']),
        ]

    configs = [
        Config(checkpoint, ablation.suffix, ablation.args)
        for checkpoint in args.checkpoints
        for ablation in ablations
    ]
    Benchmark(
        configs, args.output,
        reference_config=configs[0] if len(args.checkpoints) == 1 else None,
        stat_filter=('AP', 'AP0.5', 'AP0.75', 'APM', 'APL'),
        stat_scale=100.0,
    ).run()


if __name__ == '__main__':
    main()



""" Cross-Covariance Image Transformer (XCiT) in PyTorch
Same as the official implementation, with some minor adaptations.
    - https://github.com/facebookresearch/xcit/blob/master/detection/backbone/xcit.py
Paper:
    - https://arxiv.org/abs/2106.09681
"""
# Copyright (c) 2015-present, Facebook, Inc.
# All rights reserved.
# Licensed under The Apache-2.0 License [see docs/LICENSE.XCIT for details]
# --------------------------------------------------------
# Refactoring, modifications to class arguments and
# loading of pretrained model weights by David Mizrahi
# --------------------------------------------------------
# pylint: skip-file

import math
from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F

try:
    from timm.models.vision_transformer import Mlp
    from timm.models.layers import DropPath, trunc_normal_, to_2tuple
except ImportError:
    pass


class PositionalEncodingFourier(nn.Module):
    """
    Positional encoding relying on a fourier kernel matching the one used in the
    "Attention is all of Need" paper. The implementation builds on DeTR code
    https://github.com/facebookresearch/detr/blob/master/models/position_encoding.py
    """

    def __init__(self, hidden_dim=32, dim=768, temperature=10000):
        super().__init__()
        self.token_projection = nn.Conv2d(hidden_dim * 2, dim, kernel_size=1)
        self.scale = 2 * math.pi
        self.temperature = temperature
        self.hidden_dim = hidden_dim
        self.dim = dim

    def forward(self, B, H, W):
        mask = torch.zeros(B, H, W).bool().to(self.token_projection.weight.device)
        not_mask = ~mask
        y_embed = not_mask.cumsum(1, dtype=torch.float32)
        x_embed = not_mask.cumsum(2, dtype=torch.float32)
        eps = 1e-6
        y_embed = y_embed / (y_embed[:, -1:, :] + eps) * self.scale
        x_embed = x_embed / (x_embed[:, :, -1:] + eps) * self.scale

        dim_t = torch.arange(self.hidden_dim, dtype=torch.float32, device=mask.device)
        dim_t = self.temperature ** (2 * (dim_t // 2) / self.hidden_dim)

        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        pos_x = torch.stack((pos_x[:, :, :, 0::2].sin(),
                             pos_x[:, :, :, 1::2].cos()), dim=4).flatten(3)
        pos_y = torch.stack((pos_y[:, :, :, 0::2].sin(),
                             pos_y[:, :, :, 1::2].cos()), dim=4).flatten(3)
        pos = torch.cat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)
        pos = self.token_projection(pos)
        return pos


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return torch.nn.Sequential(
        nn.Conv2d(
            in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False
        ),
        nn.BatchNorm2d(out_planes)
    )


class ConvPatchEmbed(nn.Module):
    """ Image to Patch Embedding using multiple convolutional layers
    """

    def __init__(self, patch_size=16, embed_dim=768):
        super().__init__()
        patch_size = to_2tuple(patch_size)
        self.patch_size = patch_size

        if patch_size[0] == 16:
            self.proj = torch.nn.Sequential(
                conv3x3(3, embed_dim // 8, 2),
                nn.GELU(),
                conv3x3(embed_dim // 8, embed_dim // 4, 2),
                nn.GELU(),
                conv3x3(embed_dim // 4, embed_dim // 2, 2),
                nn.GELU(),
                conv3x3(embed_dim // 2, embed_dim, 2),
            )
        elif patch_size[0] == 8:
            self.proj = torch.nn.Sequential(
                conv3x3(3, embed_dim // 4, 2),
                nn.GELU(),
                conv3x3(embed_dim // 4, embed_dim // 2, 2),
                nn.GELU(),
                conv3x3(embed_dim // 2, embed_dim, 2),
            )
        else:
            raise("For convolutional projection, patch size has to be in [8, 16]")

    def forward(self, x):
        x = self.proj(x)
        Hp, Wp = x.shape[2], x.shape[3]
        x = x.flatten(2).transpose(1, 2)

        return x, (Hp, Wp)


class LPI(nn.Module):
    """
    Local Patch Interaction module that allows explicit communication between tokens in 3x3 windows
    to augment the implicit communcation performed by the block diagonal scatter attention.
    Implemented using 2 layers of separable 3x3 convolutions with GeLU and BatchNorm2d
    """

    def __init__(self, in_features, out_features=None, act_layer=nn.GELU, kernel_size=3):
        super().__init__()
        out_features = out_features or in_features

        padding = kernel_size // 2

        self.conv1 = torch.nn.Conv2d(in_features, out_features, kernel_size=kernel_size,
                                     padding=padding, groups=out_features)
        self.act = act_layer()
        self.bn = nn.BatchNorm2d(in_features)
        self.conv2 = torch.nn.Conv2d(in_features, out_features, kernel_size=kernel_size,
                                     padding=padding, groups=out_features)

    def forward(self, x, H, W):
        B, N, C = x.shape
        x = x.permute(0, 2, 1).reshape(B, C, H, W)
        x = self.conv1(x)
        x = self.act(x)
        x = self.bn(x)
        x = self.conv2(x)
        x = x.reshape(B, C, N).permute(0, 2, 1)

        return x


# Not used for dense feature maps
class ClassAttention(nn.Module):
    """Class Attention Layer as in CaiT https://arxiv.org/abs/2103.17239
    """

    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads)
        qkv = qkv.permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)

        qc = q[:, :, 0:1]   # CLS token
        attn_cls = (qc * k).sum(dim=-1) * self.scale
        attn_cls = attn_cls.softmax(dim=-1)
        attn_cls = self.attn_drop(attn_cls)

        cls_tkn = (attn_cls.unsqueeze(2) @ v).transpose(1, 2).reshape(B, 1, C)
        cls_tkn = self.proj(cls_tkn)
        x = torch.cat([self.proj_drop(cls_tkn), x[:, 1:]], dim=1)
        return x


# Not used for dense feature maps
class ClassAttentionBlock(nn.Module):
    """Class Attention Layer as in CaiT https://arxiv.org/abs/2103.17239
    """

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0.,
                 attn_drop=0., drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, eta=None,
                 tokens_norm=False):
        super().__init__()
        self.norm1 = norm_layer(dim)

        self.attn = ClassAttention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop,
            proj_drop=drop
        )

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer,
                       drop=drop)

        if eta is not None:     # LayerScale Initialization (no layerscale when None)
            self.gamma1 = nn.Parameter(eta * torch.ones(dim), requires_grad=True)
            self.gamma2 = nn.Parameter(eta * torch.ones(dim), requires_grad=True)
        else:
            self.gamma1, self.gamma2 = 1.0, 1.0

        self.tokens_norm = tokens_norm

    def forward(self, x, H, W, mask=None):
        x = x + self.drop_path(self.gamma1 * self.attn(self.norm1(x)))
        if self.tokens_norm:
            x = self.norm2(x)
        else:
            x[:, 0:1] = self.norm2(x[:, 0:1])

        x_res = x
        cls_token = x[:, 0:1]
        cls_token = self.gamma2 * self.mlp(cls_token)
        x = torch.cat([cls_token, x[:, 1:]], dim=1)
        x = x_res + self.drop_path(x)
        return x


class XCA(nn.Module):
    """ Cross-Covariance Attention (XCA) operation where the channels are updated using a weighted
     sum. The weights are obtained from the (softmax normalized) Cross-covariance
    matrix (Q^T K \\in d_h \\times d_h)
    """

    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        self.temperature = nn.Parameter(torch.ones(num_heads, 1, 1))

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads)
        qkv = qkv.permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)

        q = q.transpose(-2, -1)
        k = k.transpose(-2, -1)
        v = v.transpose(-2, -1)

        q = torch.nn.functional.normalize(q, dim=-1)
        k = torch.nn.functional.normalize(k, dim=-1)

        attn = (q @ k.transpose(-2, -1)) * self.temperature
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).permute(0, 3, 1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'temperature'}


class XCABlock(nn.Module):
    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0.,
                 attn_drop=0., drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, eta=None):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = XCA(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop,
            proj_drop=drop
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)

        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer,
                       drop=drop)

        self.norm3 = norm_layer(dim)
        self.local_mp = LPI(in_features=dim, act_layer=act_layer)

        self.gamma1 = nn.Parameter(eta * torch.ones(dim), requires_grad=True)
        self.gamma2 = nn.Parameter(eta * torch.ones(dim), requires_grad=True)
        self.gamma3 = nn.Parameter(eta * torch.ones(dim), requires_grad=True)

    def forward(self, x, H, W):
        x = x + self.drop_path(self.gamma1 * self.attn(self.norm1(x)))
        x = x + self.drop_path(self.gamma3 * self.local_mp(self.norm3(x), H, W))
        x = x + self.drop_path(self.gamma2 * self.mlp(self.norm2(x)))
        return x


class XCiT(nn.Module):
    """ XCiT for dense tasks
    Based on timm and DeiT code bases
    https://github.com/rwightman/pytorch-image-models/tree/master/timm
    https://github.com/facebookresearch/deit/
    """

    def __init__(self, patch_size=16, embed_dim=768,
                 depth=12, num_heads=12, mlp_ratio=4., qkv_bias=True, qk_scale=None,
                 drop_rate=0., attn_drop_rate=0., drop_path_rate=0., norm_layer=None,
                 cls_attn_layers=2, use_pos=True, eta=None, tokens_norm=False, use_fpn=False,
                 out_indices=[3, 5, 7, 11]):
        """
        Args:
            patch_size (int, tuple): patch size
            embed_dim (int): embedding dimension
            depth (int): depth of transformer
            num_heads (int): number of attention heads
            mlp_ratio (int): ratio of mlp hidden dim to embedding dim
            qkv_bias (bool): enable bias for qkv if True
            qk_scale (float): override default qk scale of head_dim ** -0.5 if set
            drop_rate (float): dropout rate
            attn_drop_rate (float): attention dropout rate
            drop_path_rate (float): stochastic depth rate
            norm_layer: (nn.Module): normalization layer
            cls_attn_layers: (int) Depth of Class attention layers (not used for dense feature maps)
            use_pos: (bool) whether to use positional encoding
            eta: (float) layerscale initialization value
            tokens_norm: (bool) Whether to normalize all tokens or just the cls_token in the CA (not used)
            use_fpn: (bool) if True, use FPN features
            out_indices: (list) Indices of layers from which FPN features are extracted
        """
        super().__init__()
        self.num_features = self.embed_dim = embed_dim
        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)

        self.patch_embed = ConvPatchEmbed(embed_dim=embed_dim, patch_size=patch_size)

        self.pos_drop = nn.Dropout(p=drop_rate)

        dpr = [drop_path_rate for i in range(depth)]
        self.blocks = nn.ModuleList([
            XCABlock(
                dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias,
                qk_scale=qk_scale, drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i],
                norm_layer=norm_layer, eta=eta)
            for i in range(depth)])

        self.pos_embeder = PositionalEncodingFourier(dim=embed_dim)
        self.use_pos = use_pos

        self.use_fpn = use_fpn
        self.out_indices = out_indices
        self.patch_size = patch_size

        if use_fpn:
            self.fpn1, self.fpn2, self.fpn3, self.fpn4 = self.init_fpn(patch_size, embed_dim)

    def init_fpn(self, patch_size, embed_dim):
        """Initializes layers for FPN if used"""

        if patch_size == 16:
            fpn1 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
                nn.BatchNorm2d(embed_dim),
                nn.GELU(),
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            )

            fpn2 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            )

            fpn3 = nn.Identity()

            fpn4 = nn.MaxPool2d(kernel_size=2, stride=2)

        elif patch_size == 8:
            fpn1 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            )

            fpn2 = nn.Identity()

            fpn3 = nn.Sequential(
                nn.MaxPool2d(kernel_size=2, stride=2),
            )

            fpn4 = nn.Sequential(
                nn.MaxPool2d(kernel_size=4, stride=4),
            )

        else:
            raise ValueError("Invalid patch size for FPN")

        return fpn1, fpn2, fpn3, fpn4

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'pos_embed', 'cls_token', 'dist_token'}

    def init_weights(self, pretrained=None, strict=False):
        """Initialize the weights in backbone.
        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
            strict (bool): whether to strictly enforce that the keys in state_dict
                match the keys returned by this module’s state_dict() function.
                Defaults to False.
        """
        def _init_weights(m):
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

        if isinstance(pretrained, str):
            self.apply(_init_weights)

            if pretrained.startswith('https'):
                checkpoint = torch.hub.load_state_dict_from_url(pretrained, map_location='cpu')
            else:
                checkpoint = torch.load(pretrained, map_location='cpu')

            checkpoint_model = checkpoint['model']
            missing_keys = self.load_state_dict(checkpoint_model, strict=strict)
            # Uncomment to print missing_keys
            # print(missing_keys)

        elif pretrained is None:
            self.apply(_init_weights)
        else:
            raise TypeError('pretrained must be a str or None')

    def extract_fpn(self, x, B, Hp, Wp):
        features = []
        for i, blk in enumerate(self.blocks):
            x = blk(x, Hp, Wp)
            if i in self.out_indices:
                xp = x.permute(0, 2, 1).reshape(B, -1, Hp, Wp)
                features.append(xp)

        ops = [self.fpn1, self.fpn2, self.fpn3, self.fpn4]
        for i in range(len(features)):
            features[i] = ops[i](features[i])

        return tuple(features)

    def forward_features(self, x):
        B, C, H, W = x.shape
        x, (Hp, Wp) = self.patch_embed(x)

        pos_encoding = self.pos_embeder(B, Hp, Wp).reshape(B, -1, x.shape[1]).permute(0, 2, 1)

        if self.use_pos:
            x = x + pos_encoding

        x = self.pos_drop(x)

        if self.use_fpn:
            return self.extract_fpn(x, B, Hp, Wp)
        else:
            for i, blk in enumerate(self.blocks):
                x = blk(x, Hp, Wp)

            # Permute and project to obtain feature map
            xp = x.permute(0, 2, 1).reshape(B, -1, Hp, Wp)
            return xp

    def forward(self, x):
        x = self.forward_features(x)
        return x


# Architectures
# Patch size 16x16 models
def xcit_nano_12_p16(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=16, embed_dim=128, depth=12, num_heads=4, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1.0, tokens_norm=False, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_nano_12_p16_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_tiny_12_p16(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=16, embed_dim=192, depth=12, num_heads=4, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1.0, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_tiny_12_p16_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_small_12_p16(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=16, embed_dim=384, depth=12, num_heads=8, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1.0, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_small_12_p16_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_tiny_24_p16(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=16, embed_dim=192, depth=24, num_heads=4, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1e-5, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_tiny_24_p16_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_small_24_p16(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=16, embed_dim=384, depth=24, num_heads=8, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1e-5, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_small_24_p16_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_medium_24_p16(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=16, embed_dim=512, depth=24, num_heads=8, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1e-5, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_medium_24_p16_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_large_24_p16(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=16, embed_dim=768, depth=24, num_heads=16, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1e-5, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_large_24_p16_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


# Patch size 8x8 models
def xcit_nano_12_p8(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=8, embed_dim=128, depth=12, num_heads=4, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1.0, tokens_norm=False, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_nano_12_p8_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_tiny_12_p8(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=8, embed_dim=192, depth=12, num_heads=4, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1.0, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_tiny_12_p8_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_small_12_p8(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=8, embed_dim=384, depth=12, num_heads=8, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1.0, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_small_12_p8_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_tiny_24_p8(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=8, embed_dim=192, depth=24, num_heads=4, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1e-5, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_tiny_24_p8_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_small_24_p8(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=8, embed_dim=384, depth=24, num_heads=8, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1e-5, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_small_24_p8_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_medium_24_p8(pretrained=True, **kwargs):
    model = XCiT(
        patch_size=8, embed_dim=512, depth=24, num_heads=8, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1e-5, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_medium_24_p8_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model


def xcit_large_24_p8(pretrained=None, **kwargs):
    model = XCiT(
        patch_size=8, embed_dim=768, depth=24, num_heads=16, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), eta=1e-5, tokens_norm=True, **kwargs)
    url = "https://dl.fbaipublicfiles.com/xcit/xcit_large_24_p8_384_dist.pth"
    model.init_weights(url if pretrained else None)
    return model



import argparse

import torch

import openpifpaf
from openpifpaf.plugins.coco.constants import (
    COCO_CATEGORIES,
    COCO_KEYPOINTS,
    COCO_PERSON_SKELETON,
    COCO_PERSON_SIGMAS,
    COCO_PERSON_SCORE_WEIGHTS,
    COCO_UPRIGHT_POSE,
    DENSER_COCO_PERSON_CONNECTIONS,
    HFLIP,
)
from openpifpaf.transforms.pair import SingleImage as S

try:
    import pycocotools.coco
    # monkey patch for Python 3 compat
    pycocotools.coco.unicode = str
except ImportError:
    pass


class CocoKpSt(openpifpaf.datasets.DataModule):
    max_shift = 30.0

    def __init__(self):
        super().__init__()

        cif = openpifpaf.headmeta.TSingleImageCif(
            'cif', 'cocokpst',
            keypoints=COCO_KEYPOINTS,
            sigmas=COCO_PERSON_SIGMAS,
            pose=COCO_UPRIGHT_POSE,
            draw_skeleton=COCO_PERSON_SKELETON,
            score_weights=COCO_PERSON_SCORE_WEIGHTS,
        )
        caf = openpifpaf.headmeta.TSingleImageCaf(
            'caf', 'cocokpst',
            keypoints=COCO_KEYPOINTS,
            sigmas=COCO_PERSON_SIGMAS,
            pose=COCO_UPRIGHT_POSE,
            skeleton=COCO_PERSON_SKELETON,
        )
        dcaf = openpifpaf.headmeta.TSingleImageCaf(
            'caf25', 'cocokpst',
            keypoints=COCO_KEYPOINTS,
            sigmas=COCO_PERSON_SIGMAS,
            pose=COCO_UPRIGHT_POSE,
            skeleton=DENSER_COCO_PERSON_CONNECTIONS,
            sparse_skeleton=COCO_PERSON_SKELETON,
            only_in_field_of_view=True,
        )
        tcaf = openpifpaf.headmeta.Tcaf(
            'tcaf', 'cocokpst',
            keypoints_single_frame=COCO_KEYPOINTS,
            sigmas_single_frame=COCO_PERSON_SIGMAS,
            pose_single_frame=COCO_UPRIGHT_POSE,
            draw_skeleton_single_frame=COCO_PERSON_SKELETON,
            only_in_field_of_view=True,
        )

        cif.upsample_stride = openpifpaf.plugins.coco.CocoKp.upsample_stride
        caf.upsample_stride = openpifpaf.plugins.coco.CocoKp.upsample_stride
        dcaf.upsample_stride = openpifpaf.plugins.coco.CocoKp.upsample_stride
        tcaf.upsample_stride = openpifpaf.plugins.coco.CocoKp.upsample_stride
        self.head_metas = ([cif, caf, dcaf, tcaf]
                           if openpifpaf.plugins.coco.CocoKp.with_dense
                           else [cif, caf, tcaf])

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module CocoKpSt')
        group.add_argument('--cocokpst-max-shift',
                           default=cls.max_shift, type=float,
                           help='max shift')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.max_shift = args.cocokpst_max_shift

    def _preprocess(self):
        bmin = openpifpaf.plugins.coco.CocoKp.bmin
        encoders = (
            openpifpaf.encoder.SingleImage(openpifpaf.encoder.Cif(self.head_metas[0], bmin=bmin)),
            openpifpaf.encoder.SingleImage(openpifpaf.encoder.Caf(self.head_metas[1], bmin=bmin)),
            openpifpaf.encoder.SingleImage(openpifpaf.encoder.Caf(self.head_metas[2], bmin=bmin)),
            openpifpaf.encoder.Tcaf(self.head_metas[3], bmin=bmin),
        ) if len(self.head_metas) == 4 else (
            openpifpaf.encoder.SingleImage(openpifpaf.encoder.Cif(self.head_metas[0], bmin=bmin)),
            openpifpaf.encoder.SingleImage(openpifpaf.encoder.Caf(self.head_metas[1], bmin=bmin)),
            openpifpaf.encoder.Tcaf(self.head_metas[2], bmin=bmin),
        )

        if not openpifpaf.plugins.coco.CocoKp.augmentation:
            return openpifpaf.transforms.Compose([
                openpifpaf.transforms.NormalizeAnnotations(),
                openpifpaf.transforms.RescaleAbsolute(openpifpaf.plugins.coco.CocoKp.square_edge),
                openpifpaf.transforms.CenterPad(openpifpaf.plugins.coco.CocoKp.square_edge),
                openpifpaf.transforms.pair.ImageToTracking(),
                S(openpifpaf.transforms.EVAL_TRANSFORM),
                openpifpaf.transforms.pair.Encoders(encoders),
            ])

        if openpifpaf.plugins.coco.CocoKp.extended_scale:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.25 * openpifpaf.plugins.coco.CocoKp.rescale_images,
                             2.0 * openpifpaf.plugins.coco.CocoKp.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))
        else:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.4 * openpifpaf.plugins.coco.CocoKp.rescale_images,
                             2.0 * openpifpaf.plugins.coco.CocoKp.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))

        return openpifpaf.transforms.Compose([
            openpifpaf.transforms.NormalizeAnnotations(),
            openpifpaf.transforms.pair.ImageToTracking(),
            openpifpaf.transforms.RandomApply(openpifpaf.transforms.pair.RandomizeOneFrame(), 0.2),
            S(openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.HFlip(COCO_KEYPOINTS, HFLIP), 0.5)),
            S(rescale_t),
            S(openpifpaf.transforms.RandomChoice(
                [openpifpaf.transforms.RotateBy90(angle_perturbation=30.0, prepad=True),
                 openpifpaf.transforms.RotateUniform(30.0)],
                [openpifpaf.plugins.coco.CocoKp.orientation_invariant, 0.4],
            )),
            openpifpaf.transforms.pair.Crop(
                openpifpaf.plugins.coco.CocoKp.square_edge, max_shift=self.max_shift),
            openpifpaf.transforms.pair.Pad(
                openpifpaf.plugins.coco.CocoKp.square_edge, max_shift=self.max_shift),
            S(openpifpaf.transforms.RandomChoice([openpifpaf.transforms.Blur(),
                                                  openpifpaf.transforms.HorizontalBlur()],
                                                 [openpifpaf.plugins.coco.CocoKp.blur / 2.0,
                                                  openpifpaf.plugins.coco.CocoKp.blur / 2.0])),
            S(openpifpaf.transforms.TRAIN_TRANSFORM),
            openpifpaf.transforms.pair.Encoders(encoders),
        ])

    def train_loader(self):
        train_data = openpifpaf.plugins.coco.CocoDataset(
            image_dir=openpifpaf.plugins.coco.CocoKp.train_image_dir,
            ann_file=openpifpaf.plugins.coco.CocoKp.train_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=openpifpaf.plugins.coco.CocoKp.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            train_data,
            batch_size=self.batch_size // 2,
            shuffle=(not openpifpaf.plugins.coco.CocoKp.debug
                     and openpifpaf.plugins.coco.CocoKp.augmentation),
            pin_memory=openpifpaf.plugins.coco.CocoKp.pin_memory,
            num_workers=self.loader_workers,
            drop_last=True,
            collate_fn=openpifpaf.datasets.collate_tracking_images_targets_meta,
        )

    def val_loader(self):
        val_data = openpifpaf.plugins.coco.CocoDataset(
            image_dir=openpifpaf.plugins.coco.CocoKp.val_image_dir,
            ann_file=openpifpaf.plugins.coco.CocoKp.val_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=openpifpaf.plugins.coco.CocoKp.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            val_data,
            batch_size=self.batch_size // 2,
            shuffle=False,
            pin_memory=openpifpaf.plugins.coco.CocoKp.pin_memory,
            num_workers=self.loader_workers,
            drop_last=True,
            collate_fn=openpifpaf.datasets.collate_tracking_images_targets_meta,
        )

    def _eval_preprocess(self):
        return openpifpaf.transforms.Compose([
            *openpifpaf.plugins.coco.CocoKp.common_eval_preprocess(),
            openpifpaf.transforms.ToAnnotations([
                openpifpaf.transforms.ToKpAnnotations(
                    COCO_CATEGORIES,
                    keypoints_by_category={1: self.head_metas[0].keypoints},
                    skeleton_by_category={1: self.head_metas[1].skeleton},
                ),
                openpifpaf.transforms.ToCrowdAnnotations(COCO_CATEGORIES),
            ]),
            openpifpaf.transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = openpifpaf.plugins.coco.CocoDataset(
            image_dir=openpifpaf.plugins.coco.CocoKp.eval_image_dir,
            ann_file=openpifpaf.plugins.coco.CocoKp.eval_annotations,
            preprocess=self._eval_preprocess(),
            annotation_filter=openpifpaf.plugins.coco.CocoKp.eval_annotation_filter,
            min_kp_anns=(openpifpaf.plugins.coco.CocoKp.min_kp_anns
                         if openpifpaf.plugins.coco.CocoKp.eval_annotation_filter
                         else 0),
            category_ids=[1] if openpifpaf.plugins.coco.CocoKp.eval_annotation_filter else [],
        )
        return torch.utils.data.DataLoader(
            eval_data,
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=openpifpaf.plugins.coco.CocoKp.pin_memory,
            num_workers=self.loader_workers,
            drop_last=False,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta,
        )

    def metrics(self):
        return [openpifpaf.metric.Coco(
            pycocotools.coco.COCO(openpifpaf.plugins.coco.CocoKp.eval_annotations),
            max_per_image=20,
            category_ids=[1],
            iou_type='keypoints',
        )]



"""Train a neural net."""

import argparse
import copy
import hashlib
import logging
import shutil
import time

import torch

from ..profiler import TorchProfiler

LOG = logging.getLogger(__name__)


class Trainer():
    epochs = None
    n_train_batches = None
    n_val_batches = None

    clip_grad_norm = 0.0
    clip_grad_value = 0.0
    log_interval = 11
    val_interval = 1

    fix_batch_norm = False
    stride_apply = 1
    ema_decay = 0.01
    train_profile = None
    distributed_reduce_loss = True

    def __init__(self, model, loss, optimizer, out, *,
                 checkpoint_shell=None,
                 lr_scheduler=None,
                 device=None,
                 model_meta_data=None):
        self.model = model
        self.loss = loss
        self.optimizer = optimizer
        self.out = out
        self.checkpoint_shell = checkpoint_shell
        self.lr_scheduler = lr_scheduler
        self.device = device
        self.model_meta_data = model_meta_data

        self.ema = None
        self.ema_restore_params = None

        self.n_clipped_grad = 0
        self.max_norm = 0.0

        if self.train_profile and (not torch.distributed.is_initialized()
                                   or torch.distributed.get_rank() == 0):
            # monkey patch to profile self.train_batch()
            self.train_batch = TorchProfiler(self.train_batch, out_name=self.train_profile)

        LOG.info({
            'type': 'config',
            'field_names': self.loss.field_names,
        })

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('trainer')
        group.add_argument('--epochs', type=int,
                           help='number of epochs to train')
        group.add_argument('--train-batches', default=None, type=int,
                           help='number of train batches')
        group.add_argument('--val-batches', default=None, type=int,
                           help='number of val batches')

        group.add_argument('--clip-grad-norm', default=cls.clip_grad_norm, type=float,
                           help='clip grad norm: specify largest change for single param')
        group.add_argument('--clip-grad-value', default=cls.clip_grad_value, type=float,
                           help='clip grad value: specify largest change for single param')
        group.add_argument('--log-interval', default=cls.log_interval, type=int,
                           help='log loss every n steps')
        group.add_argument('--val-interval', default=cls.val_interval, type=int,
                           help='validation run every n epochs')

        group.add_argument('--stride-apply', default=cls.stride_apply, type=int,
                           help='apply and reset gradients every n batches')
        assert not cls.fix_batch_norm
        group.add_argument('--fix-batch-norm',
                           default=False, const=True, type=int, nargs='?',
                           help='fix batch norm running statistics (optionally specify epoch)')
        group.add_argument('--ema', default=cls.ema_decay, type=float,
                           help='ema decay constant')
        group.add_argument('--profile', default=cls.train_profile,
                           help='enables profiling. specify path for chrome tracing file')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.epochs = args.epochs
        cls.n_train_batches = args.train_batches
        cls.n_val_batches = args.val_batches

        cls.clip_grad_norm = args.clip_grad_norm
        cls.clip_grad_value = args.clip_grad_value
        cls.log_interval = args.log_interval
        cls.val_interval = args.val_interval

        cls.fix_batch_norm = args.fix_batch_norm
        cls.stride_apply = args.stride_apply
        cls.ema_decay = args.ema
        cls.train_profile = args.profile

    def lr(self):
        for param_group in self.optimizer.param_groups:
            return param_group['lr']

    def step_ema(self):
        if self.ema is None:
            return

        for p, ema_p in zip(self.model.parameters(), self.ema):
            ema_p.mul_(1.0 - self.ema_decay).add_(p.data, alpha=self.ema_decay)

    def apply_ema(self):
        if self.ema is None:
            return

        LOG.info('applying ema')
        self.ema_restore_params = copy.deepcopy(
            [p.data for p in self.model.parameters()])
        for p, ema_p in zip(self.model.parameters(), self.ema):
            p.data.copy_(ema_p)

    def ema_restore(self):
        if self.ema_restore_params is None:
            return

        LOG.info('restoring params from before ema')
        for p, ema_p in zip(self.model.parameters(), self.ema_restore_params):
            p.data.copy_(ema_p)
        self.ema_restore_params = None

    def loop(self,
             train_scenes: torch.utils.data.DataLoader,
             val_scenes: torch.utils.data.DataLoader,
             start_epoch=0):
        if start_epoch >= self.epochs:
            raise Exception('start epoch ({}) >= total epochs ({})'
                            ''.format(start_epoch, self.epochs))

        if self.lr_scheduler is not None:
            assert self.lr_scheduler.last_epoch == start_epoch * len(train_scenes)

        for epoch in range(start_epoch, self.epochs):
            if epoch == 0:
                self.write_model(0, final=False)
            if hasattr(train_scenes.sampler, 'set_epoch'):
                train_scenes.sampler.set_epoch(epoch)
            if hasattr(val_scenes.sampler, 'set_epoch'):
                val_scenes.sampler.set_epoch(epoch)

            self.train(train_scenes, epoch)

            if (epoch + 1) % self.val_interval == 0 \
               or epoch + 1 == self.epochs:
                self.write_model(epoch + 1, epoch + 1 == self.epochs)
                self.val(val_scenes, epoch + 1)

    # pylint: disable=method-hidden,too-many-branches,too-many-statements
    def train_batch(self, data, targets, apply_gradients=True):
        if self.device.type != 'cpu':
            assert data.is_pinned(), 'input data must be pinned'
            if targets[0] is not None:
                assert targets[0].is_pinned(), 'input targets must be pinned'
            with torch.autograd.profiler.record_function('to-device'):
                data = data.to(self.device, non_blocking=True)
                targets = [head.to(self.device, non_blocking=True)
                           if head is not None else None
                           for head in targets]

        # train encoder
        with torch.autograd.profiler.record_function('model'):
            outputs = self.model(data, head_mask=[t is not None for t in targets])
            if self.train_profile and self.device.type != 'cpu':
                torch.cuda.synchronize()
        with torch.autograd.profiler.record_function('loss'):
            loss, head_losses = self.loss(outputs, targets)
            if self.train_profile and self.device.type != 'cpu':
                torch.cuda.synchronize()
        if loss is not None:
            with torch.autograd.profiler.record_function('backward'):
                loss.backward()
                if self.train_profile and self.device.type != 'cpu':
                    torch.cuda.synchronize()
        if self.clip_grad_norm:
            with torch.autograd.profiler.record_function('clip-grad-norm'):
                max_norm = self.clip_grad_norm / self.lr()
                total_norm = torch.nn.utils.clip_grad_norm_(
                    self.model.parameters(), max_norm, norm_type=float('inf'))
                self.max_norm = max(float(total_norm), self.max_norm)
                if total_norm > max_norm:
                    self.n_clipped_grad += 1
                    print('CLIPPED GRAD NORM: total norm before clip: {}, max norm: {}'
                          ''.format(total_norm, max_norm))
                if self.train_profile and self.device.type != 'cpu':
                    torch.cuda.synchronize()
        if self.clip_grad_value:
            with torch.autograd.profiler.record_function('clip-grad-value'):
                torch.nn.utils.clip_grad_value_(self.model.parameters(), self.clip_grad_value)
                if self.train_profile and self.device.type != 'cpu':
                    torch.cuda.synchronize()
        if apply_gradients:
            with torch.autograd.profiler.record_function('step'):
                self.optimizer.step()
                self.optimizer.zero_grad()
                if self.train_profile and self.device.type != 'cpu':
                    torch.cuda.synchronize()
            with torch.autograd.profiler.record_function('ema'):
                self.step_ema()
                if self.train_profile and self.device.type != 'cpu':
                    torch.cuda.synchronize()

        with torch.inference_mode():
            with torch.autograd.profiler.record_function('reduce-losses'):
                loss = self.reduce_loss(loss)
                head_losses = self.reduce_loss(head_losses)
                if self.train_profile and self.device.type != 'cpu':
                    torch.cuda.synchronize()

        return (
            float(loss.item()) if loss is not None else None,
            [float(l.item()) if l is not None else None
             for l in head_losses],
        )

    @classmethod
    def reduce_loss(cls, loss):
        if not cls.distributed_reduce_loss:
            return loss
        if loss is None:
            return loss
        if not torch.distributed.is_initialized():
            return loss

        if isinstance(loss, (list, tuple)):
            return [cls.reduce_loss(l) for l in loss]

        # average loss from all processes
        torch.distributed.reduce(loss, 0)
        if torch.distributed.get_rank() == 0:
            loss = loss / torch.distributed.get_world_size()
        return loss

    def val_batch(self, data, targets):
        if self.device:
            data = data.to(self.device, non_blocking=True)
            targets = [head.to(self.device, non_blocking=True)
                       if head is not None else None
                       for head in targets]

        with torch.inference_mode():
            outputs = self.model(data)
            loss, head_losses = self.loss(outputs, targets)
            loss = self.reduce_loss(loss)
            head_losses = self.reduce_loss(head_losses)

        return (
            float(loss.item()) if loss is not None else None,
            [float(l.item()) if l is not None else None
             for l in head_losses],
        )

    # pylint: disable=too-many-branches
    def train(self, scenes, epoch):
        start_time = time.time()
        self.model.train()
        if self.fix_batch_norm is True \
           or (self.fix_batch_norm is not False and self.fix_batch_norm <= epoch):
            LOG.info('fix batchnorm')
            for m in self.model.modules():
                if isinstance(m, (torch.nn.BatchNorm1d, torch.nn.BatchNorm2d)):
                    LOG.debug('eval mode for: %s', m)
                    m.eval()

        self.ema_restore()
        self.ema = None

        epoch_loss = 0.0
        head_epoch_losses = None
        head_epoch_counts = None
        last_batch_end = time.time()
        self.optimizer.zero_grad()
        for batch_idx, (data, target, _) in enumerate(scenes):
            preprocess_time = time.time() - last_batch_end

            batch_start = time.time()
            apply_gradients = batch_idx % self.stride_apply == 0
            loss, head_losses = self.train_batch(data, target, apply_gradients)

            # update epoch accumulates
            if loss is not None:
                epoch_loss += loss
            if head_epoch_losses is None:
                head_epoch_losses = [0.0 for _ in head_losses]
                head_epoch_counts = [0 for _ in head_losses]
            for i, head_loss in enumerate(head_losses):
                if head_loss is None:
                    continue
                head_epoch_losses[i] += head_loss
                head_epoch_counts[i] += 1

            batch_time = time.time() - batch_start

            # write training loss
            if batch_idx % self.log_interval == 0:
                batch_info = {
                    'type': 'train',
                    'epoch': epoch, 'batch': batch_idx, 'n_batches': len(scenes),
                    'time': round(batch_time, 3),
                    'data_time': round(preprocess_time, 3),
                    'lr': round(self.lr(), 8),
                    'loss': round(loss, 3) if loss is not None else None,
                    'head_losses': [round(l, 3) if l is not None else None
                                    for l in head_losses],
                }
                if hasattr(self.loss, 'batch_meta'):
                    batch_info.update(self.loss.batch_meta())
                LOG.info(batch_info)

            # initialize ema
            if self.ema is None and self.ema_decay:
                self.ema = copy.deepcopy([p.data for p in self.model.parameters()])

            # update learning rate
            if self.lr_scheduler is not None:
                self.lr_scheduler.step()

            if self.n_train_batches and batch_idx + 1 >= self.n_train_batches:
                break

            last_batch_end = time.time()

        self.apply_ema()
        LOG.info({
            'type': 'train-epoch',
            'epoch': epoch + 1,
            'loss': round(epoch_loss / len(scenes), 5),
            'head_losses': [round(l / max(1, c), 5)
                            for l, c in zip(head_epoch_losses, head_epoch_counts)],
            'time': round(time.time() - start_time, 1),
            'n_clipped_grad': self.n_clipped_grad,
            'max_norm': self.max_norm,
        })
        self.n_clipped_grad = 0
        self.max_norm = 0.0

    def val(self, scenes, epoch):
        start_time = time.time()

        # Train mode implies outputs are for losses, so have to use it here.
        self.model.train()
        if self.fix_batch_norm is True \
           or (self.fix_batch_norm is not False and self.fix_batch_norm <= epoch - 1):
            LOG.info('fix batchnorm')
            for m in self.model.modules():
                if isinstance(m, (torch.nn.BatchNorm1d, torch.nn.BatchNorm2d)):
                    LOG.debug('eval mode for: %s', m)
                    m.eval()

        epoch_loss = 0.0
        head_epoch_losses = None
        head_epoch_counts = None
        for batch_idx, (data, target, _) in enumerate(scenes):
            loss, head_losses = self.val_batch(data, target)

            # update epoch accumulates
            if loss is not None:
                epoch_loss += loss
            if head_epoch_losses is None:
                head_epoch_losses = [0.0 for _ in head_losses]
                head_epoch_counts = [0 for _ in head_losses]
            for i, head_loss in enumerate(head_losses):
                if head_loss is None:
                    continue
                head_epoch_losses[i] += head_loss
                head_epoch_counts[i] += 1

            if self.n_val_batches and batch_idx + 1 >= self.n_val_batches:
                break

        eval_time = time.time() - start_time

        LOG.info({
            'type': 'val-epoch',
            'epoch': epoch,
            'loss': round(epoch_loss / len(scenes), 5),
            'head_losses': [round(l / max(1, c), 5)
                            for l, c in zip(head_epoch_losses, head_epoch_counts)],
            'time': round(eval_time, 1),
        })

    def write_model(self, epoch, final=True):
        if torch.distributed.is_initialized() and torch.distributed.get_rank() != 0:
            return

        model_to_save = self.model
        if self.checkpoint_shell is not None:
            model = self.model if not hasattr(self.model, 'module') else self.model.module
            self.checkpoint_shell.load_state_dict(model.state_dict())
            model_to_save = self.checkpoint_shell

        filename = '{}.epoch{:03d}'.format(self.out, epoch)
        LOG.debug('about to write model')
        torch.save({
            'model': model_to_save,
            'epoch': epoch,
            'meta': self.model_meta_data,
        }, filename)
        LOG.info('model written: %s', filename)

        if final:
            sha256_hash = hashlib.sha256()
            with open(filename, 'rb') as f:
                for byte_block in iter(lambda: f.read(8192), b''):
                    sha256_hash.update(byte_block)
            file_hash = sha256_hash.hexdigest()
            outname, _, outext = self.out.rpartition('.')
            final_filename = '{}-{}.{}'.format(outname, file_hash[:8], outext)
            shutil.copyfile(filename, final_filename)



"""Export a checkpoint as an ONNX model.

Applies onnx utilities to improve the exported model and
also tries to simplify the model with onnx-simplifier.

https://github.com/onnx/onnx/blob/master/docs/PythonAPIOverview.md
https://github.com/daquexian/onnx-simplifier
"""

import argparse
import logging
import shutil

import torch

import openpifpaf

try:
    import onnx
except ImportError:
    onnx = None

try:
    import onnxsim
except ImportError:
    onnxsim = None

LOG = logging.getLogger(__name__)


def image_size_warning(basenet_stride, input_w, input_h):
    if input_w % basenet_stride != 1:
        LOG.warning(
            'input width (%d) should be a multiple of basenet '
            'stride (%d) + 1: closest are %d and %d',
            input_w, basenet_stride,
            (input_w - 1) // basenet_stride * basenet_stride + 1,
            ((input_w - 1) // basenet_stride + 1) * basenet_stride + 1,
        )

    if input_h % basenet_stride != 1:
        LOG.warning(
            'input height (%d) should be a multiple of basenet '
            'stride (%d) + 1: closest are %d and %d',
            input_h, basenet_stride,
            (input_h - 1) // basenet_stride * basenet_stride + 1,
            ((input_h - 1) // basenet_stride + 1) * basenet_stride + 1,
        )


def apply(model, outfile, verbose=True, input_w=129, input_h=97):
    image_size_warning(model.base_net.stride, input_w, input_h)

    # configure
    openpifpaf.network.heads.CompositeField3.inplace_ops = False
    openpifpaf.network.heads.CompositeField4.inplace_ops = False

    dummy_input = torch.randn(1, 3, input_h, input_w)

    torch.onnx.export(
        model, dummy_input, outfile, verbose=verbose,
        input_names=['input_batch'], output_names=['cif', 'caf'],
        # keep_initializers_as_inputs=True,
        opset_version=11,
        do_constant_folding=True,
        dynamic_axes={
            'input_batch': {0: 'dynamic'},
            'cif': {0: 'dynamic'},
            'caf': {0: 'dynamic'},
        },
    )


def check(modelfile):
    model = onnx.load(modelfile)
    onnx.checker.check_model(model)


def simplify(infile, outfile=None, input_w=129, input_h=97):
    if outfile is None:
        assert infile.endswith('.onnx')
        outfile = infile
        infile = infile.replace('.onnx', '.unsimplified.onnx')
        shutil.copyfile(outfile, infile)

    simplified_model, check_ok = onnxsim.simplify(
        infile,
        input_shapes={'input_batch': [1, 3, input_h, input_w]},
        check_n=3,
        perform_optimization=False,
    )
    assert check_ok
    onnx.save(simplified_model, outfile)


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def main():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.export_onnx',
        description=__doc__,
        formatter_class=CustomFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=openpifpaf.__version__))

    openpifpaf.network.Factory.cli(parser)

    parser.add_argument('--outfile', default='openpifpaf-resnet50.onnx')
    parser.add_argument('--simplify', dest='simplify', default=False, action='store_true')
    parser.add_argument('--check', dest='check', default=False, action='store_true')
    parser.add_argument('--input-width', type=int, default=129)
    parser.add_argument('--input-height', type=int, default=97)
    args = parser.parse_args()

    openpifpaf.network.Factory.configure(args)

    model, _ = openpifpaf.network.Factory().factory()

    apply(model, args.outfile, input_w=args.input_width, input_h=args.input_height)
    if args.simplify:
        simplify(args.outfile, input_w=args.input_width, input_h=args.input_height)
    if args.check:
        check(args.outfile)


if __name__ == '__main__':
    main()



import dataclasses
import logging
from typing import ClassVar, List, Tuple

from .. import headmeta
from .. import visualizer as visualizer_module
from .annrescaler import TrackingAnnRescaler
from .caf import CafGenerator

LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class Tcaf:
    """Tracking Composite Association Field."""

    meta: headmeta.Tcaf
    rescaler: TrackingAnnRescaler = None
    v_threshold: int = 0
    bmin: float = 0.1
    visualizer: visualizer_module.Tcaf = None
    fill_plan: List[Tuple[int, int, int]] = None

    min_size: ClassVar[int] = 3
    fixed_size: ClassVar[bool] = True
    aspect_ratio: ClassVar[float] = 0.0
    padding: ClassVar[int] = 10

    def __post_init__(self):
        if self.rescaler is None:
            self.rescaler = TrackingAnnRescaler(self.meta.stride, self.meta.pose)

        if self.visualizer is None:
            self.visualizer = visualizer_module.Tcaf(self.meta)

        if self.fill_plan is None:
            self.fill_plan = [
                (caf_i, joint1i - 1, joint2i - 1)
                for caf_i, (joint1i, joint2i) in enumerate(self.meta.skeleton)
            ]

    def __call__(self, images, all_anns, metas):
        return CafGenerator(self)(images[0], all_anns, metas)



import argparse
import logging
import time

import numpy as np
try:
    import scipy.optimize
except ImportError:
    scipy = None  # pylint: disable=invalid-name

from .. import headmeta
from .cifcaf import CifCaf
from .track_annotation import TrackAnnotation
from .track_base import TrackBase
from . import pose_distance

LOG = logging.getLogger(__name__)


class PoseSimilarity(TrackBase):
    distance_type = pose_distance.Euclidean

    def __init__(self, cif_meta: headmeta.Cif, caf_meta: headmeta.Caf, *, pose_generator=None):
        super().__init__()
        self.cif_meta = cif_meta
        self.caf_meta = caf_meta

        # prefer decoders with more keypoints and associations
        self.priority = -10.0
        self.priority += cif_meta.n_fields / 1000.0
        self.priority += caf_meta.n_fields / 1000.0

        self.distance_function = self.distance_type()
        self.distance_function.valid_keypoints = [
            i
            for i, kp in enumerate(cif_meta.keypoints)
            if kp not in (('left_ear', 'right_ear') if cif_meta.dataset == 'posetrack2018' else [])
        ]
        LOG.debug('valid keypoints = %s', self.distance_function.valid_keypoints)
        self.distance_function.sigmas = np.asarray(cif_meta.sigmas)

        self.pose_generator = pose_generator or CifCaf([cif_meta], [caf_meta])

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('PoseSimilarity')
        assert cls.distance_type == pose_distance.Euclidean
        group.add_argument('--posesimilarity-distance', default='euclidean',
                           choices=('crafted', 'euclidean', 'euclidean4', 'oks'))
        group.add_argument('--posesimilarity-oks-inflate',
                           default=pose_distance.Oks.inflate, type=float)

    @classmethod
    def configure(cls, args: argparse.Namespace):
        if args.posesimilarity_distance == 'euclidean':
            cls.distance_type = pose_distance.Euclidean
        elif args.posesimilarity_distance == 'euclidean4':
            cls.distance_type = lambda _: pose_distance.Euclidean(track_frames=[-1, -4, -8, -12])
        elif args.posesimilarity_distance == 'oks':
            cls.distance_type = pose_distance.Oks
        elif args.posesimilarity_distance == 'crafted':
            cls.distance_type = pose_distance.Crafted
        else:
            raise Exception('distance function type not known')

        pose_distance.Oks.inflate = args.posesimilarity_oks_inflate

    @classmethod
    def factory(cls, head_metas):
        if len(head_metas) < 2:
            return []
        return [
            cls(cif_meta, caf_meta)
            for cif_meta, caf_meta
            in zip(head_metas, head_metas[1:])
            if (isinstance(cif_meta, (headmeta.TSingleImageCif, headmeta.Cif))
                and isinstance(caf_meta, (headmeta.TSingleImageCaf, headmeta.Caf)))
        ]

    def __call__(self, fields, *, initial_annotations=None):
        self.frame_number += 1
        start = time.perf_counter()

        self.prune_active(self.frame_number)

        pose_start = time.perf_counter()
        pose_annotations = self.pose_generator(fields)
        LOG.debug('pose time = %.3fs', time.perf_counter() - pose_start)

        cost_start = time.perf_counter()
        cost = np.full((len(self.active) * 2, len(pose_annotations)), 1000.0)
        for track_i, track in enumerate(self.active):
            for pose_i, pose in enumerate(pose_annotations):
                cost[track_i, pose_i] = self.distance_function(
                    self.frame_number, pose, track, self.track_is_good(track, self.frame_number))

                # option to loose track (e.g. occlusion)
                cost[track_i + len(self.active), pose_i] = 100.0
        LOG.debug('cost time = %.3fs', time.perf_counter() - cost_start)

        if scipy is None:
            raise Exception('scipy not found. Run "pip3 install scipy" to install.')
        track_indices, pose_indices = scipy.optimize.linear_sum_assignment(cost)
        matched_poses = set()
        for track_i, pose_i in zip(track_indices, pose_indices):
            # was track lost?
            if track_i >= len(self.active):
                continue

            pose = pose_annotations[pose_i]
            track = self.active[track_i]

            track.add(self.frame_number, pose)
            matched_poses.add(pose)

        for new_pose in pose_annotations:
            if new_pose in matched_poses:
                continue
            self.active.append(TrackAnnotation().add(self.frame_number, new_pose))

        # tag ignore regions TODO
        # if self.gt_anns:
        #     self.tag_ignore_region(self.frame_number, self.gt_anns)

        # pruning lost tracks
        self.active = [t for t in self.active if self.track_is_viable(t, self.frame_number)]

        LOG.info('active tracks = %d, good = %d, track ids = %s',
                 len(self.active),
                 len([t for t in self.active if self.track_is_good(t, self.frame_number)]),
                 [self.simplified_track_id_map.get(t.id_, t.id_)
                  for t in self.active])

        if self.track_visualizer:
            self.track_visualizer.predicted(
                self.frame_number,
                [t for t in self.active if self.track_is_good(t, self.frame_number)],
            )

        LOG.debug('track time: %.3fs', time.perf_counter() - start)
        return self.annotations(self.frame_number)



from .preprocess import Preprocess


class Encoders(Preprocess):
    """Preprocess operation that runs encoders."""
    def __init__(self, encoders):
        self.encoders = encoders

    def __call__(self, image, anns, meta):
        anns = [enc(image, anns, meta) for enc in self.encoders]
        meta['head_indices'] = [enc.meta.head_index for enc in self.encoders]
        return image, anns, meta



import logging

from .annrescaler import AnnRescaler
from .caf import Caf
from .cif import Cif

LOG = logging.getLogger(__name__)


def cli(parser):
    group = parser.add_argument_group('encoders')
    group.add_argument('--cif-side-length', default=Cif.side_length, type=int,
                       help='side length of the CIF field')
    group.add_argument('--caf-min-size', default=Caf.min_size, type=int,
                       help='min side length of the CAF field')
    group.add_argument('--caf-fixed-size', default=Caf.fixed_size, action='store_true',
                       help='fixed caf size')
    group.add_argument('--caf-aspect-ratio', default=Caf.aspect_ratio, type=float,
                       help='CAF width relative to its length')
    assert AnnRescaler.suppress_selfhidden
    group.add_argument('--encoder-no-suppress-selfhidden',
                       dest='encoder_suppress_selfhidden',
                       default=True, action='store_false',
                       help='[experimental]')
    assert not AnnRescaler.suppress_invisible
    group.add_argument('--encoder-suppress-invisible',
                       default=False, action='store_true',
                       help='[experimental]')
    assert not AnnRescaler.suppress_collision
    group.add_argument('--encoder-suppress-collision',
                       default=False, action='store_true',
                       help='[experimental]')


def configure(args):
    # configure CIF
    Cif.side_length = args.cif_side_length

    # configure CAF
    Caf.min_size = args.caf_min_size
    Caf.fixed_size = args.caf_fixed_size
    Caf.aspect_ratio = args.caf_aspect_ratio

    # configure AnnRescaler
    AnnRescaler.suppress_selfhidden = args.encoder_suppress_selfhidden
    AnnRescaler.suppress_invisible = args.encoder_suppress_invisible
    AnnRescaler.suppress_collision = args.encoder_suppress_collision



import logging

from .base import Base
from .. import show

LOG = logging.getLogger(__name__)


class Seeds(Base):
    """Visualize seeds."""

    def __init__(self, *, stride=1):
        super().__init__('seeds')
        self.stride = stride

    def predicted(self, seeds):
        """Seeds are: confidence, field_index, x, y, ..."""
        field_indices = self.indices()
        if not field_indices:
            return

        with self.image_canvas(self.processed_image()) as ax:
            show.white_screen(ax)
            for f in field_indices:
                if f == -1:
                    f_seeds = seeds
                else:
                    f_seeds = [s for s in seeds if f == s[1]]
                x = [s[2] * self.stride for s in f_seeds]
                y = [s[3] * self.stride for s in f_seeds]
                ax.plot(x, y, 'o')
                if f in self.indices('confidence', with_all=False):
                    c = [s[0] for s in f_seeds]
                    for xx, yy, cc in zip(x, y, c):
                        ax.text(xx, yy, '{:.2f}'.format(cc))
